package dk.rsd.dcc;

import dk.rsd.dcc.configuration.IntegrationTestConfiguration;
import dk.rsd.dcc.util.*;
import dk.sosi.seal.xml.XmlUtil;
import dk.sundhedsdatastyrelsen.dcc.wsdl.DGWSFault;
import dk.sundhedsdatastyrelsen.dcc.wsdl.DccNTSServicePort;
import dk.sundhedsdatastyrelsen.dcc.xsd.DccRequestType;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringRunner;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;

import java.net.HttpURLConnection;
import java.net.URL;
import java.util.Optional;
import java.util.Properties;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

@RunWith(SpringRunner.class)
@ContextConfiguration(classes = {IntegrationTestConfiguration.class})
public class DccServletIntegrationsTest {
    @Value("${dcc.host}")
    String dccHost;

    @Autowired
    private DccNTSServicePort dccService;

    @Autowired
    @Qualifier("3Millisec")
    private DccNTSServicePort dccService3millisec;

    @Autowired
    @Qualifier("sosigateway")
    private DccNTSServicePort dccNTSWithSOSIGateWayServicePort;


    @Test
    public void testRegularRequest() throws Exception {

        String payload = "payload";

        DccRequestType request = new DccRequestType();
        request.setDccRequstUrl(payload);

        String result = dccService.dccCall(request);
        assertNotNull(result);
    }

    @Test
    public void testMissingContentTypeRequest() throws Exception {
        // DCC service will automatically set content type, which is why it's not used here.
        URL url = new URL("http://localhost:8080/decoupling");
        HttpURLConnection httpConn = (HttpURLConnection) url.openConnection();
        httpConn.setRequestProperty("SOAPAction", "http://sosi.dk/its#handleRequestRequest");
        httpConn.setRequestMethod("GET");

        int responseCode = httpConn.getResponseCode();

        assertEquals(500, responseCode);
        Document errorDoc = XmlUtil.readXml(new Properties(), new InputSource(httpConn.getErrorStream()), false);
        Node faultstringNode = errorDoc.getDocumentElement().getElementsByTagName("faultstring").item(0);
        assertNotNull(faultstringNode);
        assertEquals("Missing Content-Type header", faultstringNode.getTextContent());
    }

    @Test
    public void testTimeoutRequest() {

        String expectedFaultcode = "dcc_timeout";
        String resultat = null;

        String payload = "payload";

        DccRequestType request = new DccRequestType();
        request.setDccRequstUrl(payload);

        try {
            resultat = dccService3millisec.dccCall(request);
        } catch (DGWSFault dgwsFault) {
            String fault = dgwsFault.getFaultInfo();
            assertEquals("Exception " + fault + " modtaget", expectedFaultcode, fault);

        } catch (Exception e) {
            e.printStackTrace();
            assertTrue("Ikke " + expectedFaultcode + " exception", false);
        }
        assertNotEquals("",resultat);
    }

    @Test
   public void testRequestWithServiceIdentifier1() throws Exception {
        String serviceIdentifier = "service-1";

        DccRequestType request = new DccRequestType();
        request.setDccRequstUrl("payload");

        DccNTSServicePort port = DccNTSServicePortFactory.buildDccNTSServicePort(dccHost, "http://nspop.dk/nts/2013/05#invoke", Optional.of(serviceIdentifier));

        String result = port.dccCall(request);
        assertNotNull(result);
    }

    @Test
    public void testRequestWithUnknownServiceIdentifier() throws Exception {
        String serviceIdentifier = "unknown-service";
        String payload = "";

        DccRequestType request = new DccRequestType();
        request.setDccRequstUrl(payload);

        DccNTSServicePort port = DccNTSServicePortFactory.buildDccNTSServicePort(dccHost, "http://nspop.dk/nts/2013/05#invoke", Optional.of(serviceIdentifier));

        try {
            port.dccCall(request);
            assertTrue(false);
        }
        catch(DGWSFault e) {
            // Deliberately empty - the call should fail.
        }
    }
}

