package dk.rsd.dcc;/*
 * SOSI-DCC - Copyright (C) 2011 National Board of e-Health (NSI)
 *
 * All source code and information supplied as part of SOSI-DCC is
 * copyright too Digital Sundhed.
 *
 * The source code has been released under a dual license - meaning you can
 * use either licensed version of the library with your code.
 *
 * It is released under the Common Public License 1.0, a copy of which can
 * be found at the link below.
 * http://www.opensource.org/licenses/cpl1.0.php
 *
 * It is released under the LGPL (GNU Lesser General Public License), either
 * version 2.1 of the License, or (at your option) any later version. A copy
 * of which can be found at the link below.
 * http://www.gnu.org/copyleft/lesser.html
 *
 * $HeadURL: https://svn.nspop.dk/svn/lakeside/sosi-dcc/tags/release-2.4.3/modules/dcc/src/main/java/dk/rsd/dcc/DccConfig.java $
 * $Id: DccConfig.java 33 2017-01-12 16:52:05Z chg $
 */

import dk.rsd.dcc.fault.AccessDeniedFault;
import dk.rsd.dcc.fault.DccFault;
import dk.rsd.dcc.routing.*;
import dk.sosi.seal.model.constants.DGWSConstants;
import dk.sosi.seal.xml.DebugErrorHandler;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.GeneralSecurityException;
import java.security.KeyStore;
import java.util.*;
import java.util.concurrent.atomic.AtomicReference;

import static java.lang.Boolean.*;

public class DccConfig {

    private static final String LOG4J_DCC_PROPERTIES = "log4j-dcc.properties";
    private static final String JBOSS_CONFIG_DIR = "jboss.server.config.dir";
    private static final String WAS_REPOSITORY_ROOT = "was.repository.root";
    private static final String CATALINA_HOME = "catalina.home";
    private static final String USER_HOME = "user.home";
    private static final String FILE_SEP = System.getProperty("file.separator");

    private static final String SCHEMA_LANGUAGE = "http://java.sun.com/xml/jaxp/properties/schemaLanguage";
    private static final String XML_SCHEMA = "http://www.w3.org/2001/XMLSchema";
    private static final String SCHEMA_SOURCE = "http://java.sun.com/xml/jaxp/properties/schemaSource";

    private static final Logger log = Logger.getLogger(DccConfig.class);

    private static final List<String> HTTP_HEADERS_SET_BY_DCC = Arrays.asList("Content-Length", "Content-Type", "SOAPAction", "X-Forwarded-For");

    private static final AtomicReference<DccConfig> instance = new AtomicReference<>(new DccConfig());
    private Set<String> whitelistedIpPrefixes;
    private Set<String> incomingHttpHeadersToPreserve;
    private String dgwsVersion;
    private String serviceIdentifierPrefix;
    private boolean nspMode;
    private String trustStoreFile;
    private String trustStorePassword;

    private final RoutingConfiguration routingConfiguration;

    public static void reload() { instance.set(new DccConfig()); }

    public static class ConfigException extends RuntimeException {
        private static final long serialVersionUID = -5889015933127210026L;
    }

    public static void initLogging() {
        String configDirectory = getConfigDirectory();
        String log4jPropertyPath = configDirectory + LOG4J_DCC_PROPERTIES;

        if (configDirectory == null || !new File(configDirectory).exists()) {
            PropertyConfigurator.configure(DccConfig.class.getResource("/" + LOG4J_DCC_PROPERTIES));
            log.warn("Configuration directory '" + configDirectory + "' not found! Loaded default log4j configuration from classpath instead.");
        } else if (new File(log4jPropertyPath).exists()) {
            PropertyConfigurator.configure(log4jPropertyPath);
            log.info("Loaded log4j configuration from " + log4jPropertyPath);
        } else {
            PropertyConfigurator.configure(DccConfig.class.getResource("/" + LOG4J_DCC_PROPERTIES));
            log.warn("'" + LOG4J_DCC_PROPERTIES + "' not found in '" + configDirectory + "'! Loaded default log4j configuration from classpath instead.");
        }
    }

    private static String getConfigDirectory() {
        String configDir = null;
        if (System.getProperty(DccConstants.PROPERTY_DCC_CONFIG_DIR) != null) {
            configDir = System.getProperty(DccConstants.PROPERTY_DCC_CONFIG_DIR + FILE_SEP);
        } else if (System.getProperty(JBOSS_CONFIG_DIR) != null) {
            configDir = System.getProperty(JBOSS_CONFIG_DIR) + FILE_SEP;
        } else if (System.getProperty(WAS_REPOSITORY_ROOT) != null) {
            // On WAS the configuration is placed in the user's home dir since
            // the contents of the was.repository.root directory are sometimes
            /// deleted by WAS on some installations :(
            configDir = System.getProperty(USER_HOME) + FILE_SEP + "sosi-dcc" + FILE_SEP;
        } else if (System.getProperty(CATALINA_HOME) != null) {
            configDir = System.getProperty(CATALINA_HOME) + FILE_SEP + "conf" + FILE_SEP;
        }
        return configDir;
    }

    private DccConfig() {
        try {
            Document doc = parseConfiguration();
            Element nspMode = (Element) doc.getDocumentElement().getElementsByTagName("NSPMode").item(0);
            loadNSPMode(nspMode);
            Element dgwsVersion = (Element) doc.getDocumentElement().getElementsByTagName("DGWSVersion").item(0);
            loadDGWSVersion(dgwsVersion);
            NodeList serviceIdentifierPrefixElementList = doc.getDocumentElement().getElementsByTagName("ServiceIdentifierPrefix");
            if(serviceIdentifierPrefixElementList.getLength() > 0) {
                loadServiceIdentifierPrefix((Element) serviceIdentifierPrefixElementList.item(0));
            }
            Element incomingHttpHeadersToPreserve = (Element) doc.getDocumentElement().getElementsByTagName("PreservedIncomingHttpHeaders").item(0);
            loadIncomingHttpHeadersToPreserve(incomingHttpHeadersToPreserve);
            Element trustStore = (Element) doc.getDocumentElement().getElementsByTagName("TrustStore").item(0);
            loadElementTrustStore(trustStore);
            Element clientAccessWhitelist = (Element) doc.getDocumentElement().getElementsByTagName("ClientAccessWhitelist").item(0);
            loadClientAccessWhitelist(clientAccessWhitelist);
            Element endpointDatabase = (Element) doc.getDocumentElement().getElementsByTagName("EndpointDatabase").item(0);
            RoutingConfigurator routingConfigurator = endpointDatabase != null
                    ? loadEndpointDatabaseConfigurator(endpointDatabase)
                    : new XMLBasedRoutingConfigurator(doc.getDocumentElement().getElementsByTagName("Endpoint"));
            routingConfiguration = routingConfigurator.getConfiguration();
            log.info("Configuration loaded ...");
        } catch (ParserConfigurationException | SAXException | IOException e) {
            log.error("Unable to load configuration", e);
            throw new ConfigException();
        }
    }

    private void loadElementTrustStore(Element trustStore) {
        if (trustStore != null) {
            this.trustStoreFile = loadNotEmptyTextChild(trustStore.getElementsByTagName("JKS_File").item(0));
            this.trustStorePassword = loadNotEmptyTextChild(trustStore.getElementsByTagName("Password").item(0));
        }
    }

    private void loadNSPMode(Element nspModeElement) {
        this.nspMode = nspModeElement != null && parseBoolean(loadNotEmptyTextChild(nspModeElement));
    }

    private void loadDGWSVersion(Element dgwsVersionElement) {
        String dgwsVersion = loadNotEmptyTextChild(dgwsVersionElement);
        if (!DGWSConstants.SUPPORTED_VERSIONS.contains(dgwsVersion)) {
            log.error("Unsupported DGWSVersion: '" + dgwsVersion + "'. Supported versions are: " + DGWSConstants.SUPPORTED_VERSIONS);
            throw new ConfigException();
        }
        this.dgwsVersion = dgwsVersion;
    }

    private void loadServiceIdentifierPrefix(Element serviceIdentifierPrefixElement) {
        this.serviceIdentifierPrefix = loadNotEmptyTextChild(serviceIdentifierPrefixElement);
    }

    private void loadIncomingHttpHeadersToPreserve(Element headersToPreserve) {
        incomingHttpHeadersToPreserve = new TreeSet<String>(String.CASE_INSENSITIVE_ORDER);
        NodeList ipPrefixNodes = headersToPreserve.getElementsByTagName("HttpHeader");
        loadNodeListIntoSet(ipPrefixNodes, incomingHttpHeadersToPreserve);
        incomingHttpHeadersToPreserve.removeAll(HTTP_HEADERS_SET_BY_DCC);
    }

    private void loadClientAccessWhitelist(Element clientAccessWhitelist) {
        this.whitelistedIpPrefixes = new HashSet<String>();
        if (clientAccessWhitelist == null) {
            log.warn("Client access whitelisting not enabled! All client IPs will be able to use the service.");
            return;
        }
        NodeList ipPrefixNodes = clientAccessWhitelist.getElementsByTagName("IP_Prefix");
        loadNodeListIntoSet(ipPrefixNodes, this.whitelistedIpPrefixes);
    }

    private void loadNodeListIntoSet(NodeList nodeList, Set<String> set) {
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node node = nodeList.item(i);
            String content = loadNotEmptyTextChild(node);
            set.add(content);
        }
    }

    private DatabaseBasedRoutingConfigurator loadEndpointDatabaseConfigurator(Element endpointDatabase) {
        String dataSourceName = loadNotEmptyTextChild(endpointDatabase.getElementsByTagName("DataSourceName").item(0));
        URL gatewayProxyURL = loadGatewayProxyURL(endpointDatabase);
        String actionsSelectStatement = loadNotEmptyTextChild(endpointDatabase.getElementsByTagName("ActionsSelectStatement").item(0));
        return new DatabaseBasedRoutingConfigurator(dataSourceName, gatewayProxyURL, actionsSelectStatement);
    }

    private URL loadGatewayProxyURL(Element endpointDatabase) {
        NodeList gatewayProxyURLNodes = endpointDatabase.getElementsByTagName("GatewayProxyURL");
        if (gatewayProxyURLNodes.getLength() > 0) {
            String proxyURLString = loadNotEmptyTextChild(gatewayProxyURLNodes.item(0));
            if (!proxyURLString.trim().equals("")) {
                try {
                    return new URL(proxyURLString);
                } catch (MalformedURLException e) {
                    log.error("Error parsing GatewayProxyURL element", e);
                    throw new ConfigException();
                }
            }
        }
        return null;
    }

    private Document parseConfiguration() throws ParserConfigurationException, SAXException, IOException {
        InputStream schemaStream = null;
        InputStream configStream = null;
        try {
            DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
            docBuilderFactory.setAttribute(SCHEMA_LANGUAGE, XML_SCHEMA);
            String schemaFile = System.getProperty(DccConstants.PROPERTY_DCC_SCHEMAFILE, DccConstants.DCC_DEFAULT_SCHEMAFILE);
            schemaStream = XmlUtil.class.getResourceAsStream("/" + schemaFile);
            docBuilderFactory.setAttribute(SCHEMA_SOURCE, schemaStream);
            docBuilderFactory.setValidating(true);

            configStream = getConfigurationAsStream();

            DocumentBuilder builder = docBuilderFactory.newDocumentBuilder();
            builder.setErrorHandler(new DebugErrorHandler(false));
            return builder.parse(configStream);
        } finally {
            if (schemaStream != null) {
                schemaStream.close();
            }
            if (configStream != null) {
                configStream.close();
            }
        }
    }

    private InputStream getConfigurationAsStream() throws IOException {
        String configFileName = System.getProperty(DccConstants.PROPERTY_DCC_CONFIGFILE, DccConstants.DCC_DEFAULT_CONFIGFILE);
        InputStream configStream;

        String configDirectory = getConfigDirectory();
        if (configDirectory != null) {
            File configFile = new File(configDirectory, configFileName);
            if (!configFile.exists()) {
                log.error("'" + configFileName + "' not found in '" + configDirectory + "'!");
                throw new ConfigException();
            }
            configStream = new FileInputStream(configFile);
        } else {
            log.info("Neither '" +
                    JBOSS_CONFIG_DIR + "', '" +
                    DccConstants.PROPERTY_DCC_CONFIG_DIR + "', '" +
                    WAS_REPOSITORY_ROOT + "' nor '" +
                    CATALINA_HOME + "' are set! Trying to load configuration from classpath ...");
            configStream = DccConfig.class.getResourceAsStream("/" + configFileName);
        }
        return configStream;
    }

    private String loadNotEmptyTextChild(Node node) throws ConfigException {
        Node child = node.getFirstChild();
        if (child == null || child.getNodeValue().trim().length() == 0) {
            log.error("Error in configuration. Element '" + node.getNodeName() + "' is empty!");
            throw new ConfigException();
        } else {
            return child.getNodeValue().trim();
        }
    }

    public static DccConfig getInstance() { return instance.get(); }

    public URL getServiceURL(String soapAction, String serviceIdentifier) throws DccFault {
        return routingConfiguration.getServiceURL(soapAction, serviceIdentifier);
    }

    public URL getProxyURL(String soapAction, String serviceIdentifier) throws DccFault {
        return routingConfiguration.getProxyURL(soapAction, serviceIdentifier);
    }

    public DecouplingInfo getDefaultDecouplingInfo(String soapAction, String serviceIdentifier) throws DccFault {
        return routingConfiguration.getDefaultDecouplingInfo(soapAction, serviceIdentifier);
    }

    public boolean doNotVerifySSLHostname(String soapAction, String serviceIdentifier) {
        return routingConfiguration.doNotVerifySSLHostname(soapAction, serviceIdentifier);
    }

    public WSAHeaderProcessing getWSAHeaderProcessing(String soapAction, String serviceIdentifier) {
        return routingConfiguration.getWSAHeaderProcessing(soapAction, serviceIdentifier);
    }

    public int getIdCardMaxAgeInMinutes(String soapAction, String serviceIdentifier) {
        return routingConfiguration.getIdCardMaxAgeInMinutes(soapAction, serviceIdentifier);
    }

    public String getDGWSVersion() {
        return dgwsVersion;
    }

    public String getServiceIdentifierPrefix() {
        return serviceIdentifierPrefix;
    }

    public void checkClientAccessWhitelist(String ipAdress) throws DccFault {
        // if no whitelist is specified in the configuration file access is unrestricted (useful for testing)
        if (whitelistedIpPrefixes.size() != 0) {
            for (Iterator<String> iterator = whitelistedIpPrefixes.iterator(); iterator.hasNext(); ) {
                String prefix = iterator.next();
                if (ipAdress.startsWith(prefix)) return;
            }
            String msg = "Client-ip '" + ipAdress + "' not in access whitelist";
            log.warn(msg);
            throw new AccessDeniedFault(msg);
        }
    }

    public boolean preserveHttpHeader(String header) {
        return incomingHttpHeadersToPreserve.contains(header);
    }

    public boolean useSLALog() {
        return nspMode;
    }

    public boolean useIsAlive() {
        return nspMode;
    }

    public KeyStore getTrustStore() throws ConfigException{
        if (trustStoreFile != null) {
            return loadTrustStore();
        } else {
            return null;
        }
    }

    private KeyStore loadTrustStore() throws ConfigException {
        try {
            KeyStore trustStore = KeyStore.getInstance(KeyStore.getDefaultType());
            String configDirectory = getConfigDirectory();
            InputStream is;
            if (configDirectory != null) {
                log.info("Trying to load truststore ...");
                is = new FileInputStream(configDirectory + this.trustStoreFile);
            } else {
                log.warn("Trying to load truststore from classpath (should only happen during unittest) ...");
                is = this.getClass().getResourceAsStream("/" + this.trustStoreFile);
                if (is == null) throw new IOException(this.trustStoreFile + " not found on classpath");
            }
            trustStore.load(is, trustStorePassword.toCharArray());
            log.info("Loaded truststore '" + this.trustStoreFile + "'");
            return trustStore;
        } catch (GeneralSecurityException | IOException e) {
            log.error("Unable to load truststore", e);
            throw new ConfigException();
        }
    }

}
