/*
 * SOSI-DCC - Copyright (C) 2011 National Board of e-Health (NSI)
 *
 * All source code and information supplied as part of SOSI-DCC is
 * copyright to National Board of e-Health.
 *
 * The source code has been released under a dual license - meaning you can
 * use either licensed version of the library with your code.
 *
 * It is released under the Common Public License 1.0, a copy of which can
 * be found at the link below.
 * http://www.opensource.org/licenses/cpl1.0.php
 *
 * It is released under the LGPL (GNU Lesser General Public License), either
 * version 2.1 of the License, or (at your option) any later version. A copy
 * of which can be found at the link below.
 * http://www.gnu.org/copyleft/lesser.html
 *
 * $HeadURL: https://svn.nspop.dk/svn/lakeside/sosi-dcc/tags/release-2.4.3/modules/dcc/src/main/java/dk/rsd/dcc/DccHandler.java $
 * $Id: DccHandler.java 33 2017-01-12 16:52:05Z chg $
 */
package dk.rsd.dcc;


import dk.rsd.dcc.routing.WSAHeaderProcessing;
import org.apache.commons.lang.StringEscapeUtils;
import org.apache.log4j.Logger;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import dk.sosi.seal.model.constants.MedComTags;
import dk.sosi.seal.model.constants.MedcomAttributes;
import dk.sosi.seal.model.constants.NameSpaces;
import dk.sosi.seal.model.constants.SAMLAttributes;
import dk.sosi.seal.model.constants.SAMLTags;
import dk.sosi.seal.model.constants.SOSIAttributes;

public class DccHandler extends DefaultHandler {

	private static final String HEADER = "Header";
	private static final String XML_HEADER = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
	private static final String DECOUPLING_NAMESPACE = "http://www.sosi.dk/2007/07/decoupling-component-1.0.xsd";
    private static final String SOSIGW_NAMESPACE = "http://sosi.dk/gw/2007.09.01";

	static Logger log = Logger.getLogger(DccHandler.class);

	private StringBuffer sb;
	private String currentElm;
	private String currentNS;
	private String lastSamlAttr;
	private StringBuffer charBuffer;

	private DccContext context;
	private boolean inSoapHeader;

	public DccHandler(DccContext context) {
		this.context = context;
		sb = new StringBuffer();
		charBuffer = new StringBuffer();
	}

	public void startDocument() throws SAXException {
		sb.append(XML_HEADER);
	}


	public void endDocument() throws SAXException {
		context.setDecoupledRequest(sb.toString());
	}

	public void startElement(String namespaceURI, String lName, String qName, Attributes attrs) throws SAXException {
        processCharacters();
        currentNS = namespaceURI;
		currentElm = lName;
		if (NameSpaces.SOAP_SCHEMA.equals(namespaceURI) && HEADER.equals(lName)) {
			inSoapHeader = true;
		}
		if (!doStripCurrentDecouplingOrWSAHeader(namespaceURI)) {
			sb.append("<" + qName);
			if (attrs != null) {
				processAttributes(namespaceURI, lName, attrs);
			}
			sb.append(">");
		}
		// Matches soap:Header
        if (NameSpaces.SOAP_SCHEMA.equals(namespaceURI) && lName.equals(HEADER)) {
			if (overwriteWSA()) {
                sb.append("<To xmlns=\"" + NameSpaces.WSA_SCHEMA + "\">" + context.getServiceURL() + "</To>");
                sb.append("<Action xmlns=\"" + NameSpaces.WSA_SCHEMA + "\">" + context.getSoapAction() + "</Action>");
            }
			int idCardMaxAgeInMinutes = DccConfig.getInstance().getIdCardMaxAgeInMinutes(context.getSoapAction(), context.getServiceIdentifier());
			if (idCardMaxAgeInMinutes > 0) {
				sb.append("<MaxAge xmlns=\"" + SOSIGW_NAMESPACE + "\">" + idCardMaxAgeInMinutes + "</MaxAge>");
			}
		}
	}

    private void processAttributes(String namespaceURI, String lName, Attributes attrs) {
		for (int i = 0; i < attrs.getLength(); i++) {
			// skip dcc namespace declaration
			if (DECOUPLING_NAMESPACE.equals(attrs.getValue(i))) continue;
			if (SAMLAttributes.NAME.equals(attrs.getQName(i)) && SAMLTags.ATTRIBUTE.equals(lName) && NameSpaces.SAML2ASSERTION_SCHEMA.equals(namespaceURI)) {
				lastSamlAttr = attrs.getValue(i);
			}
			sb.append(" " +  attrs.getQName(i) + "=\"" + StringEscapeUtils.escapeXml(attrs.getValue(i)) + "\"");
		}
	}

	public void endElement(String namespaceURI, String lName, String qName) throws SAXException {
		processCharacters();
		if (!doStripCurrentDecouplingOrWSAHeader(namespaceURI)) {
			sb.append("</" + qName + ">");
		}
		currentElm = null;
		currentNS = null;
		lastSamlAttr = null;
		if (NameSpaces.SOAP_SCHEMA.equals(namespaceURI) && HEADER.equals(lName)) {
			inSoapHeader = false;
		}
	}

	public void characters(char buffer[], int offset, int length) throws SAXException {
		// We need to accumulate the characters since the SAX parser is not required to report all
		// character data in a single call to characters(). Usually a SAX parser only uses
		// fixed sized buffer to hold the data that is parsed.
		charBuffer.append(buffer, offset, length);
	}

	private void processCharacters() throws SAXException {
		String string = charBuffer.toString();
		charBuffer.setLength(0);

		if (isInDecouplingHeader(currentNS)) {
            try {
                if (DecouplingInfo.TIMEOUT.equals(currentElm)) context.setTimeout(Integer.parseInt(string));
            } catch (NumberFormatException e) {
                throw new SAXException("Unable to parse decoupling timeout as a number: " + string);
            }
        } else if (isInWSAddressingHeader(currentNS) && ! retainWSA()) {
            // do nothing
		} else {
			if (MedcomAttributes.IT_SYSTEM_NAME.equals(lastSamlAttr)) {
				context.setItSystemName(string);
			} else if (SOSIAttributes.IDCARD_VERSION.equals(lastSamlAttr)) {
				context.setDGWSVersion(string);
			} else if (NameSpaces.MEDCOM_SCHEMA.equals(currentNS)){
				if (MedComTags.MESSAGE_ID.equals(currentElm)) context.setMessageID(string);
				else if (MedComTags.FLOW_ID.equals(currentElm)) context.setFlowID(string);
			}
			sb.append(StringEscapeUtils.escapeXml(string));
		}
	}

    private boolean doStripCurrentDecouplingOrWSAHeader(String namespaceURI) {
        return isInDecouplingHeader(namespaceURI) || (isInWSAddressingHeader(namespaceURI) && ! retainWSA());
    }

    private boolean isInDecouplingHeader(String namespaceURI) {
        return inSoapHeader && DECOUPLING_NAMESPACE.equals(namespaceURI);
    }

    private boolean isInWSAddressingHeader(String namespaceURI) {
        return inSoapHeader && ( NameSpaces.WSA_SCHEMA.equals(namespaceURI) || NameSpaces.WSA_1_0_SCHEMA.equals(namespaceURI));
    }

    private boolean overwriteWSA() {
        return WSAHeaderProcessing.WSA_OVERWRITE.equals(context.getWSAHeaderProcessing());
    }

    private boolean retainWSA() {
        return WSAHeaderProcessing.WSA_PASSTHROUGH.equals(context.getWSAHeaderProcessing());
    }

}