/*
 *
 *  * The MIT License
 *  *
 *  * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *  *
 *  * Copyright (C) 2015 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *  *
 *  * Permission is hereby granted, free of charge, to any person obtaining a copy of
 *  * this software and associated documentation files (the "Software"), to deal in
 *  * the Software without restriction, including without limitation the rights to
 *  * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 *  * of the Software, and to permit persons to whom the Software is furnished to do
 *  * so, subject to the following conditions:
 *  *
 *  * The above copyright notice and this permission notice shall be included in all
 *  * copies or substantial portions of the Software.
 *  *
 *  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *  * SOFTWARE.
 *  *
 *  * $HeadURL: https://svn.nspop.dk/svn/lakeside/sosi-dcc/tags/release-2.4.3/modules/dcc/src/main/java/dk/rsd/dcc/DccParser.java $
 *  * $Id: DccParser.java 33 2017-01-12 16:52:05Z chg $
 *
 */

package dk.rsd.dcc;

import dk.rsd.dcc.fault.GeneralFault;
import dk.rsd.dcc.fault.SyntaxFault;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.xml.sax.SAXException;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

/**
 * Created by chg on 10/08/15.
 */
public class DccParser {

    private static final byte[] CRLF_BYTES = new byte[]{13, 10};
    private static final byte[] DOUBLE_CRLF_BYTES = new byte[]{13, 10, 13, 10};


    private static Logger log = Logger.getLogger(DccParser.class);

    public void parse(InputStream inputStream, DccContext context) throws IOException, SyntaxFault, GeneralFault {
        String contentType = context.getContentType();
        if (contentType == null) {
            throw new SyntaxFault("Missing Content-Type header");
        }
        if (contentType.startsWith("multipart/related")) {

            String boundaryHeader = StringUtils.substringBetween(contentType, "boundary=\"", "\";");
            if (boundaryHeader == null) {
                throw new SyntaxFault("Missing 'boundary' parameter in Content-Type header");
            }
            String boundary = "--" + boundaryHeader;

            byte[] requestAsBytes = DccUtil.consumeStreamToByteArray(inputStream);

            int firstMimeBoundaryPosition = KMPMatch.indexOf(requestAsBytes, (boundary).getBytes(), 0);
            if (firstMimeBoundaryPosition != 0) {
                throw new SyntaxFault("Payload in 'multipart/related' request does not begin with expected MIME-boundary: " + boundary);
            }

            int firstDoubleCRLFPosition = KMPMatch.indexOf(requestAsBytes, DOUBLE_CRLF_BYTES, 0); // According to RFC822 fields and messagebody must be separated by two CRLFs
            if (firstDoubleCRLFPosition == -1) {
                throw new SyntaxFault("Payload in 'multipart/related' request does not contain expected double CRLF");
            }

            int secondMimeBoundaryPosition = KMPMatch.indexOf(requestAsBytes, boundary.getBytes(), firstDoubleCRLFPosition);
            if (secondMimeBoundaryPosition == -1) {
                throw new SyntaxFault("Payload in 'multipart/related' request is missing MIME-boundary: " + boundary);
            }

            int startSoapBodyPosition = firstDoubleCRLFPosition + DOUBLE_CRLF_BYTES.length;

            String decoupledSoapEnvelope = saxParseSoapEnvelope(new ByteArrayInputStream(requestAsBytes, startSoapBodyPosition, secondMimeBoundaryPosition - startSoapBodyPosition), context);

            ByteArrayOutputStream baos = new ByteArrayOutputStream(requestAsBytes.length + 200); // Should be large enough too hold decoupled request
            baos.write(requestAsBytes, 0, startSoapBodyPosition);                                                       // Leading MIME-boundary and fields for first MIME part
            baos.write(decoupledSoapEnvelope.getBytes("UTF-8"));                                                        // SOAP envelope
            baos.write(CRLF_BYTES);                                                                                     // Line break before second MIME-boundary
            baos.write(requestAsBytes, secondMimeBoundaryPosition, requestAsBytes.length - secondMimeBoundaryPosition); // Second MIME-boundary and rest of message

            context.setDecoupledRequest(null);
            context.setDecoupledRequestAsBytes(baos.toByteArray());

        } else {
            saxParseSoapEnvelope(inputStream, context);
        }
    }

    private String saxParseSoapEnvelope(InputStream inputStream, DccContext context) throws IOException, SyntaxFault, GeneralFault {
        try {
            SAXParserFactory factory = SAXParserFactory.newInstance();
            factory.setNamespaceAware(true);
            factory.setFeature("http://xml.org/sax/features/namespaces", true);
            factory.setFeature("http://xml.org/sax/features/namespace-prefixes", true);

            SAXParser parser = factory.newSAXParser();
            parser.parse(inputStream, new DccHandler(context));
            return context.getDecoupledRequestAsString();
        } catch (SAXException e) {
            log.info("Error parsing request in DCC", e);
            throw new SyntaxFault(e);
        } catch (ParserConfigurationException e) {
            log.error("Configuration error in DccServlet", e);
            throw new GeneralFault(e);
        }
    }

    /**
     * Knuth-Morris-Pratt Algorithm for Pattern Matching
     */
    private static class KMPMatch {
        /**
         * Finds the first occurrence of the pattern in the text.
         */
        public static int indexOf(byte[] data, byte[] pattern, int offset) {
            int[] failure = computeFailure(pattern);

            int j = 0;
            if (data.length == 0) return -1;

            for (int i = offset; i < data.length; i++) {
                while (j > 0 && pattern[j] != data[i]) {
                    j = failure[j - 1];
                }
                if (pattern[j] == data[i]) { j++; }
                if (j == pattern.length) {
                    return i - pattern.length + 1;
                }
            }
            return -1;
        }

        /**
         * Computes the failure function using a boot-strapping process,
         * where the pattern is matched against itself.
         */
        private static int[] computeFailure(byte[] pattern) {
            int[] failure = new int[pattern.length];

            int j = 0;
            for (int i = 1; i < pattern.length; i++) {
                while (j > 0 && pattern[j] != pattern[i]) {
                    j = failure[j - 1];
                }
                if (pattern[j] == pattern[i]) {
                    j++;
                }
                failure[i] = j;
            }

            return failure;
        }
    }

}
