/*
 * SOSI-DCC - Copyright (C) 2011 National Board of e-Health (NSI)
 *
 * All source code and information supplied as part of SOSI-DCC is
 * copyright to National Board of e-Health.
 *
 * The source code has been released under a dual license - meaning you can
 * use either licensed version of the library with your code.
 *
 * It is released under the Common Public License 1.0, a copy of which can
 * be found at the link below.
 * http://www.opensource.org/licenses/cpl1.0.php
 *
 * It is released under the LGPL (GNU Lesser General Public License), either
 * version 2.1 of the License, or (at your option) any later version. A copy
 * of which can be found at the link below.
 * http://www.gnu.org/copyleft/lesser.html
 *
 * $HeadURL: https://svn.nspop.dk/svn/lakeside/sosi-dcc/tags/release-2.4.3/modules/dcc/src/main/java/dk/rsd/dcc/DccServlet.java $
 * $Id: DccServlet.java 33 2017-01-12 16:52:05Z chg $
 */
package dk.rsd.dcc;

import dk.rsd.dcc.dispatch.Dispatcher;
import dk.rsd.dcc.dispatch.DispatcherFactory;
import dk.rsd.dcc.dispatch.impl.DispatcherFactoryImpl;
import dk.rsd.dcc.dispatch.ssl.SSLDispatcherContext;
import dk.rsd.dcc.fault.DccFault;
import dk.rsd.dcc.fault.GeneralFault;
import dk.rsd.dcc.fault.SyntaxFault;
import dk.sdsd.dcc.utils.SLALogUtil;
import dk.sdsd.nsp.slalog.api.SLALogConfig;
import dk.sdsd.nsp.slalog.api.SLALogException;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sdsd.nsp.slalog.impl.BackendInstantiator;
import org.apache.log4j.Logger;

import javax.servlet.ServletException;
import javax.servlet.UnavailableException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.GeneralSecurityException;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Properties;

import static dk.rsd.dcc.DccConfig.*;

public class DccServlet extends HttpServlet {

	private static final String SLA_LOG_APP_NAME = "NSP DCC";
	private static final String SLA_LOG_APP_SHORT_NAME = "sosidcc";
	private static final long serialVersionUID = 2198208635561717293L;
	private static final Logger log = Logger.getLogger(DccServlet.class);
    private DispatcherFactory dispatcherFactory = new DispatcherFactoryImpl();

    public void init() throws ServletException {
		super.init();
		initLogging();
		log.info("Starting DCC...");

		try {
            initSLALog();
            SSLDispatcherContext.init(getInstance().getTrustStore());
		} catch (ConfigException | IllegalStateException | GeneralSecurityException e) {
			throw new UnavailableException(e.getMessage());
		}
    }

    private void initSLALog() {
        SLALogConfig slaLogConfig;
        if (getInstance().useSLALog()) {
            slaLogConfig = new SLALogConfig(SLA_LOG_APP_NAME, SLA_LOG_APP_SHORT_NAME);
        } else {
            Properties properties = new Properties();
            properties.put(BackendInstantiator.NSPSLALOG_LOG4J_CONFIGURE_KEY, false);
            slaLogConfig = new SLALogConfig(properties);
        }
        SLALogUtil.setSLALogger(slaLogConfig.getSLALogger());
    }

    public void destroy() {
		super.destroy();
		log.info("Shutting down DCC ...");
	}

    // For unit-testing purposes
    /*pp*/ void setDispatcherFactory(DispatcherFactory dispatcherFactory) {
        this.dispatcherFactory = dispatcherFactory;
    }

	protected void doGet(HttpServletRequest req, HttpServletResponse resp) throws IOException {
		doPost(req, resp);
	}

	protected void doPost(HttpServletRequest req, HttpServletResponse resp) throws IOException {
		log.debug("Content-Type: " + req.getContentType());

		DccContext context = new DccContext();

        SLALogItem slaLogItem = SLALogUtil.getSLALogger().createLogItem("DccServlet.doPost", "101");
        SLALogUtil.setSLALogItem(slaLogItem);
        if (req.getContentLength() != -1) {
            slaLogItem.setRequestSize(req.getContentLength());
        }

        String ip = retrieveClientIP(req);
        context.setClientIP(ip);
        slaLogItem.setClientIP(ip);

        try {
			getInstance().checkClientAccessWhitelist(ip);
			initContext(req, context);
			parseRequest(req.getInputStream(), context);
			retrieveDecouplingInfo(context);

            SLALogUtil.setSLALogFromDccContext(slaLogItem, context);

            Dispatcher dispatcher = dispatcherFactory.createDispatcher();

			InputStream inputStream = dispatcher.dispatch(context);
			doResponse(resp, context, inputStream);
			slaLogItem.setCallResultOk();
		} catch (DccFault fault) {
			log.debug(fault);
			doErrorOrStatusResponse(fault, context, resp);
            slaLogItem.setCallResultError(fault.getMessage());
		} catch (Exception e) {
			slaLogItem.setCallResultError(e.getMessage());
			log.error("Caught Non-DccFault",e);
		} finally {
            try {
                slaLogItem.store();
            } catch (SLALogException e) {
                log.error("Cannot store SLA Log Item", e);
            }
            SLALogUtil.setSLALogItem(null);
        }
	}

    private String retrieveClientIP(HttpServletRequest req) {
        final Enumeration<String> xForwardedForHeaders = req.getHeaders("X-Forwarded-For");
        String xForwardedFor = null;
        while (xForwardedForHeaders.hasMoreElements()) {
            xForwardedFor = xForwardedForHeaders.nextElement();
        }

        return xForwardedFor != null ? xForwardedFor : req.getRemoteAddr();
    }

    // visible for testing purpose
    /* pp */ void initContext(HttpServletRequest req, DccContext context) throws DccFault {
		String soapAction = retrieveSoapAction(req);
		context.setSoapAction(soapAction);

        String serviceIdentifier = retrieveServiceIdentifier(req);
        context.setServiceIdentifier(serviceIdentifier);

        URL serviceURL = getInstance().getServiceURL(soapAction, serviceIdentifier);
        serviceURL = patchURL(serviceURL, req);
        context.setServiceURL(serviceURL);
        log.debug("serviceUrl: " + serviceURL);

        URL proxyURL = getInstance().getProxyURL(soapAction, serviceIdentifier);
        proxyURL = patchURL(proxyURL, req);
        context.setProxyURL(proxyURL);
        log.debug("proxyUrl: " + proxyURL);

        context.setDoNotVerifySSLHostname(getInstance().doNotVerifySSLHostname(soapAction, serviceIdentifier));
        context.setWSAHeaderProcessing(getInstance().getWSAHeaderProcessing(soapAction, serviceIdentifier));

        procesHttpHeaders(req, context);

        context.setContentType(req.getHeader("Content-Type"));
    }

    private void procesHttpHeaders(HttpServletRequest req, DccContext context) {
        Enumeration headerNames = req.getHeaderNames();
        while (headerNames.hasMoreElements()) {
            String headerName = (String) headerNames.nextElement();
            if (getInstance().preserveHttpHeader(headerName)) {
                Enumeration headers = req.getHeaders(headerName);
                while (headers.hasMoreElements()) {
                    context.addPreservedHttpHeader(headerName, (String) headers.nextElement());
                }
            }
        }
    }

    private URL patchURL(URL url, HttpServletRequest req) throws GeneralFault {
        if (url == null) return null;
        String urlString = url.toString();
        if (urlString.endsWith("/*")) {
            try {
                StringBuilder builder = new StringBuilder(urlString.substring(0, urlString.length() - 2));
                if (req.getPathInfo() != null) {
                    PathInfoComponents components = parsePathInfo(req.getPathInfo());
                    builder.append(components.urlSuffix);
                }
                if (req.getQueryString() != null && req.getQueryString().trim().length() != 0) {
                    builder.append("?");
                    builder.append(req.getQueryString());
                }
                return new URL(builder.toString());
            } catch (MalformedURLException e) {
                throw new GeneralFault("Unable to handle URL", e);
            }
        } else {
            return url;
        }
    }

    private String retrieveSoapAction(HttpServletRequest req) throws SyntaxFault {
		String soapAction = req.getHeader("SOAPAction");
		log.debug("SOAPAction: " + soapAction);
		if (soapAction == null) {
			String msg = "No SOAPAction specified in http header!";
			log.info(msg);
			throw new SyntaxFault(msg);
		}
		soapAction = soapAction.replaceAll("\"", "").replaceAll("'", "").trim();
		return soapAction;
	}

    private String retrieveServiceIdentifier(HttpServletRequest req) throws SyntaxFault {
        PathInfoComponents components = parsePathInfo(req.getPathInfo());
        return components.serviceIdentifier;
    }

    private class PathInfoComponents {
        String serviceIdentifier;
        String urlSuffix;

        PathInfoComponents(String serviceIdentifier, String urlSuffix) {
            this.serviceIdentifier = serviceIdentifier;
            this.urlSuffix = urlSuffix;
        }
    }

    private PathInfoComponents parsePathInfo(final String pathInfo) {
        // The pathInfo string may look like '/service/foo/bar', where the '/service/foo'-part specifies the ServiceIdentifier 'foo',
        // and '/bar' is just a url-suffix that may need to be carried along when forwarding the request.
        // Split the string into these two components, retaining leading and trailing slashes.
        // NB: The ServiceIdentifierPrefix-attribute must be present for ServiceIdentifiers to be detected at all.

        String normalizedPathInfo = pathInfo;
        if(normalizedPathInfo == null) {
            normalizedPathInfo = "";
        }

        if(normalizedPathInfo.startsWith("/")) {
            normalizedPathInfo = normalizedPathInfo.substring(1);
        }

        String[] fragments = normalizedPathInfo.split("/");
        String serviceIdentifierPrefix = getInstance().getServiceIdentifierPrefix();
        String serviceIdentifier = null;
        String urlSuffix = null;
        if(serviceIdentifierPrefix == null || fragments.length <= 1 || !fragments[0].equals(serviceIdentifierPrefix)) {
            urlSuffix = String.join("/", fragments);
        }
        else {
            serviceIdentifier = fragments[1];
            urlSuffix = String.join("/", Arrays.copyOfRange(fragments, 2, fragments.length));
        }

        if(pathInfo != null && pathInfo.startsWith("/") && !urlSuffix.startsWith("/")) {
            urlSuffix = "/" + urlSuffix;
        }
        if(pathInfo != null && pathInfo.endsWith("/") && !urlSuffix.endsWith("/")) {
            urlSuffix += "/";
        }

        return new PathInfoComponents(serviceIdentifier, urlSuffix);
    }

    private void retrieveDecouplingInfo(DccContext context) throws DccFault {
		DecouplingInfo info = context.getDecouplingInfo();
		if (info == null) {
			info = getInstance().getDefaultDecouplingInfo(context.getSoapAction(), context.getServiceIdentifier());
			context.setTimeout(info.getTimeout());
		} else {
			DecouplingValidator.validate(info);
		}
		log.debug("decouplingInfo: " + info);
	}

	private void parseRequest(InputStream inputStream, DccContext context) throws DccFault, IOException {
		if (log.isDebugEnabled()) {
			inputStream = logIncomingRequest(inputStream);
			log.debug("Parsing request ...");
		}
        new DccParser().parse(inputStream, context);
        if (log.isDebugEnabled()) {
            log.debug("Decoupled request: " + context.getDecoupledRequestAsString());
        }
    }

    private static InputStream logIncomingRequest(InputStream is) throws IOException {
		byte[] requestBytes = DccUtil.consumeStreamToByteArray(is);
		log.debug("Incoming request: " + new String(requestBytes, "UTF-8"));
		return new ByteArrayInputStream(requestBytes);
	}

	private void doResponse(HttpServletResponse resp, DccContext context, InputStream inputStream) throws DccFault, IOException {
		resp.setContentType(context.getResponseContentType());
		int responseContentLength = context.getResponseContentLength();
		resp.setContentLength(responseContentLength);
		resp.setStatus(context.getResponseStatusCode());
		SLALogItem slaLogItem = SLALogUtil.getSLALogItem();
        if (context.isResponseGzipEncoded()) {
            slaLogItem.addCallParameter("decoupling-response-gzipped", "" + context.isResponseGzipEncoded());
        }
        if (responseContentLength > 0) {
            slaLogItem.setReplySize(responseContentLength);
        }
        if (context.getResponseContentEncoding() != null) {
			resp.setHeader("Content-Encoding", context.getResponseContentEncoding());
		}
		long bytesCopied = DccUtil.streamResponseFromService(inputStream, resp.getOutputStream(), false, context.isResponseGzipEncoded());
		if (bytesCopied > 0 && responseContentLength <= 0) {
			slaLogItem.setReplySize(bytesCopied);
		}
		log.debug("Sending synchronous response ...");
	}

	private void doErrorOrStatusResponse(DccFault fault, DccContext context, HttpServletResponse resp) throws IOException {

		try {
			String replyStr = DccUtil.createErrorReply(context, fault);
			log.debug("Error/Status response from dcc: " + replyStr);

			resp.setStatus(HttpURLConnection.HTTP_INTERNAL_ERROR);
			resp.setContentType("text/xml; charset=utf-8");
			byte[] bytes = replyStr.getBytes("UTF-8");
			resp.setContentLength(bytes.length);
			resp.getOutputStream().write(bytes);
			SLALogUtil.getSLALogItem().setReplySize(bytes.length);
		} catch (IOException e) {
			log.error("Error while sending error or status response", e);
			throw e;
		}
	}
}

