/*
 * SOSI-DCC - Copyright (C) 2011 National Board of e-Health (NSI)
 *
 * All source code and information supplied as part of SOSI-DCC is
 * copyright to National Board of e-Health.
 *
 * The source code has been released under a dual license - meaning you can
 * use either licensed version of the library with your code.
 *
 * It is released under the Common Public License 1.0, a copy of which can
 * be found at the link below.
 * http://www.opensource.org/licenses/cpl1.0.php
 *
 * It is released under the LGPL (GNU Lesser General Public License), either
 * version 2.1 of the License, or (at your option) any later version. A copy
 * of which can be found at the link below.
 * http://www.gnu.org/copyleft/lesser.html
 *
 * $HeadURL: https://svn.nspop.dk/svn/lakeside/sosi-dcc/tags/release-2.4.3/modules/dcc/src/main/java/dk/rsd/dcc/dispatch/DccServiceCall.java $
 * $Id: DccServiceCall.java 33 2017-01-12 16:52:05Z chg $
 */
package dk.rsd.dcc.dispatch;

import dk.rsd.dcc.DccContext;
import dk.rsd.dcc.dispatch.ssl.SSLDispatcherContext;
import dk.sdsd.dcc.utils.SLALogUtil;
import dk.sdsd.nsp.slalog.api.SLALogException;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import org.apache.log4j.Logger;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLSocketFactory;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.ProtocolException;
import java.util.Map;

public class DccServiceCall implements DccWork {

	private static final long serialVersionUID = 3564747098207659781L;

	private final static Logger log = Logger.getLogger(DccServiceCall.class);

	private final DccContext context;
	private HttpURLConnection httpConn;
	private OutputStream out;
	private InputStream in;
	private ProxyIOException proxyIOException;

	public DccServiceCall(DccContext context) {
		this.context = context;
	}

	public void run() {
		log.debug("Calling service at "+ context.getProxyURL() + " ... ");
        SLALogItem slaLogItem = SLALogUtil.getSLALogger().createLogItem("DccServiceCall.run", "110");
        SLALogUtil.setSLALogFromDccContext(slaLogItem, context);
        slaLogItem.setTargetEndpoint(context.getProxyURL().toString());
        try {
			httpConn = (HttpURLConnection) context.getProxyURL().openConnection();

            handleSSL();

            byte[] bytes = context.getBytesToSend();

            setRequestProperties(context, httpConn, bytes);

            slaLogItem.setRequestSize(bytes.length);
            slaLogItem.setMessageId(context.getMessageID());
            slaLogItem.setClientIP(context.getClientIP());

            out = httpConn.getOutputStream();
			out.write(bytes);

			String contentType = httpConn.getContentType();
			context.setResponseContentType(contentType);

			int contentLength = httpConn.getContentLength();
			context.setResponseContentLength(contentLength);

			slaLogItem.setReplySize(contentLength);

			int responseCode = httpConn.getResponseCode();
			log.debug("HTTP status from service: " + responseCode);
			context.setResponseStatusCode(responseCode);

			String contentEncoding = httpConn.getContentEncoding();
			if (contentEncoding != null) {
				context.setResponseContentEncoding(contentEncoding);
				log.debug("Content-encoding in response: " + contentEncoding);
			}

			in =  responseCode >= 500 ? httpConn.getErrorStream() : httpConn.getInputStream();

            if (responseCode >= 500) {
                slaLogItem.setCallResultError("Error code " + responseCode + " returned from call.");
            } else {
                slaLogItem.setCallResultOk();
            }
        } catch (IOException e) {
			proxyIOException = new ProxyIOException(e);
			slaLogItem.setCallResultError(e.getMessage());
		} finally {
			if (out != null) try {out.close();} catch (IOException e) {/* ignore */}
            try {
                slaLogItem.store();
            } catch (SLALogException e) {
                log.error("Cannot store SLA Log Item", e);
            }
		}
	}

    private void handleSSL() {
        if (httpConn instanceof HttpsURLConnection) {
            SSLSocketFactory socketFactory = SSLDispatcherContext.getSSLContext().getSocketFactory();
            HttpsURLConnection httpsConn = (HttpsURLConnection) this.httpConn;
            httpsConn.setSSLSocketFactory(socketFactory);
            if (this.context.doNotVerifySSLHostname()) {
                httpsConn.setHostnameVerifier(new HostnameVerifier() {
                    @Override
                    public boolean verify(String s, SSLSession sslSession) {
                        return true;
                    }
                });
            }
        }
    }

    private void setRequestProperties(DccContext context, HttpURLConnection httpConn, byte[] bytes) throws ProtocolException {
		httpConn.setRequestProperty("Content-Length", String.valueOf(bytes.length));
        if (context.getContentType() != null) {
            httpConn.setRequestProperty("Content-Type", context.getContentType());
        } else {
            httpConn.setRequestProperty("Content-Type", "text/xml; charset=utf-8");
        }
		httpConn.setRequestProperty("SOAPAction", quote(context.getSoapAction()));
        httpConn.setRequestProperty("X-Forwarded-For", context.getClientIP());
		httpConn.setRequestMethod("POST");
		httpConn.setDoOutput(true);
		httpConn.setDoInput(true);
        for (Map.Entry<String, String> headerEntry : context.getPreservedHttpHeaders()) {
            httpConn.setRequestProperty(headerEntry.getKey(), headerEntry.getValue());
            log.debug("'" + headerEntry.getKey() + "' set to '" + headerEntry.getValue() + "'");
        }
	}

    private String quote(String string) {
        return "\"" + string + "\"";
    }

    public InputStream getResponseStream() throws ProxyIOException {
		if (proxyIOException != null) throw proxyIOException;
		return in;
	}

	public void release() {
		log.debug("Closing stream from: " + context.getProxyURL());
		if (httpConn != null) {
			httpConn.disconnect();
		}
	}

	public boolean isDaemon() {
		return false;
	}

}
