/*
 *
 *  * The MIT License
 *  *
 *  * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *  *
 *  * Copyright (C) 2014 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *  *
 *  * Permission is hereby granted, free of charge, to any person obtaining a copy of
 *  * this software and associated documentation files (the "Software"), to deal in
 *  * the Software without restriction, including without limitation the rights to
 *  * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 *  * of the Software, and to permit persons to whom the Software is furnished to do
 *  * so, subject to the following conditions:
 *  *
 *  * The above copyright notice and this permission notice shall be included in all
 *  * copies or substantial portions of the Software.
 *  *
 *  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *  * SOFTWARE.
 *  *
 *  * $HeadURL: https://svn.nspop.dk/svn/lakeside/sosi-dcc/tags/release-2.4.3/modules/dcc/src/main/java/dk/rsd/dcc/routing/DatabaseBasedRoutingConfigurator.java $
 *  * $Id: DatabaseBasedRoutingConfigurator.java 33 2017-01-12 16:52:05Z chg $
 *
 */

package dk.rsd.dcc.routing;

import dk.rsd.dcc.DecouplingInfo;
import org.apache.log4j.Logger;
import org.springframework.dao.DataAccessException;
import org.springframework.jdbc.core.JdbcTemplate;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.EnumSet;
import java.util.List;

import static dk.rsd.dcc.DccConfig.ConfigException;

public class DatabaseBasedRoutingConfigurator implements RoutingConfigurator {

    private static final Logger log = Logger.getLogger(DatabaseBasedRoutingConfigurator.class);

    private final JdbcTemplate jdbcTemplate;
    private final URL gatewayProxyURL;
    private final String actionsSelectStatement;

    public DatabaseBasedRoutingConfigurator(String dataSourceName, URL gatewayProxyURL, String actionsSelectStatement) throws ConfigException{
        this.gatewayProxyURL = gatewayProxyURL;
        this.actionsSelectStatement = actionsSelectStatement;
        try {
            Context ctx = new InitialContext();
            Object lookedUpObject = ctx.lookup(dataSourceName);
            if (lookedUpObject instanceof DataSource) {
                DataSource ds = (DataSource) lookedUpObject;
                jdbcTemplate = new JdbcTemplate(ds);
            } else {
                log.error("Error looking up dataSource under name '" + dataSourceName + "'. Expected javax.sql.DataSource but got " + lookedUpObject.getClass());
                throw new ConfigException();
            }
        } catch (NamingException e) {
            log.error("Error looking up dataSource under name '" + dataSourceName + "'", e);
            throw new ConfigException();
        }

    }

    public RoutingConfiguration getConfiguration() throws ConfigException {
        try {
            List<Route> routes = jdbcTemplate.query(actionsSelectStatement, (rs, rowNum) -> {
                Route route = new Route();
                route.setSoapAction(rs.getString("SOAP_ACTION"));
                route.setServiceIdentifier(rs.getString("SERVICE_IDENTIFIER"));
                route.setServiceURL(parseAsURL(rs.getString("SERVICE_URL")));

                DecouplingInfo decouplingInfo = new DecouplingInfo();
                decouplingInfo.setTimeout(rs.getInt("TIMEOUT_MILLIS"));
                route.setDecouplingInfo(decouplingInfo);

                boolean useProxy = rs.getBoolean("USE_PROXY");

                if (useProxy && gatewayProxyURL != null) {
                    route.setProxyURL(gatewayProxyURL);

                    int idCardMaxAge = rs.getInt("IDCARD_MAXAGE_MINS");
                    if (idCardMaxAge > 0) {
                        route.setIdCardMaxAge(idCardMaxAge);
                    }
                }

                route.setDoNotVerifySSLHostname(rs.getBoolean("DO_NOT_VERIFY_SSL_HOSTNAME"));
                route.setWsaHeaderProcessing(parseWSAHeaderProcessing(rs.getString("WSA_HEADERS_PROCESSING")));

                return route;
            });

            return new RoutingConfiguration(routes);

        } catch (DataAccessException e) {
            log.error("Error reading configuration from database", e);
            throw new ConfigException();
        }
    }

    private URL parseAsURL(String urlString) {
        try {
            return urlString != null ? new URL(urlString) : null;
        } catch (MalformedURLException e) {
            log.error("Could not parse '" + urlString + "' as URL", e);
            throw new ConfigException();
        }
    }

    private WSAHeaderProcessing parseWSAHeaderProcessing(String wsaHeaderProcessingString) {
        if (wsaHeaderProcessingString == null) return null;
        try {
            return WSAHeaderProcessing.valueOf(wsaHeaderProcessingString);
        } catch (IllegalArgumentException e) {
            log.error("Illegal WSA_HEADERS_PROCESSING value '" + wsaHeaderProcessingString + "'. Allowed values are " + EnumSet.allOf(WSAHeaderProcessing.class) + " or 'null'.");
            throw new ConfigException();
        }
    }

}
