/*
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2014 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $HeadURL: https://svn.nspop.dk/svn/lakeside/sosi-dcc/tags/release-2.4.3/modules/dcc/src/main/java/dk/rsd/dcc/routing/XMLBasedRoutingConfigurator.java $
 * $Id: XMLBasedRoutingConfigurator.java 33 2017-01-12 16:52:05Z chg $
 */
package dk.rsd.dcc.routing;

import dk.rsd.dcc.fault.DccFault;
import dk.rsd.dcc.DecouplingInfo;
import dk.rsd.dcc.DecouplingValidator;
import org.apache.log4j.Logger;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.LinkedList;
import java.util.List;

import static dk.rsd.dcc.DccConfig.*;

public class XMLBasedRoutingConfigurator implements RoutingConfigurator {

    private static final Logger log = Logger.getLogger(XMLBasedRoutingConfigurator.class);

    private final List<Route> routes = new LinkedList<>();

    public XMLBasedRoutingConfigurator(NodeList endpoints) { loadEndpoints(endpoints); }

    public RoutingConfiguration getConfiguration() { return new RoutingConfiguration(routes); }

    private void loadEndpoints(NodeList endpoints) {
        for (int i = 0; i < endpoints.getLength(); i++) {
            Element endpoint = (Element) endpoints.item(i);
            String actionPrefix = endpoint.getAttribute("ActionPrefix");
            try {
                URL url = new URL(endpoint.getAttribute("URL"));
                String proxyURLString = endpoint.getAttribute("ProxyURL");
                URL proxyUrl = null;
                if (!proxyURLString.isEmpty()) {
                    proxyUrl = new URL(proxyURLString);
                }
                String idcardMaxAgeString = endpoint.getAttribute("IdcardMaxAgeMinutes");
                int idcardMaxAge = -1;
                if (!idcardMaxAgeString.isEmpty()) {
                    idcardMaxAge = Integer.parseInt(idcardMaxAgeString);
                }
                boolean doNotVerifySSLHostname = Boolean.parseBoolean(endpoint.getAttribute("DoNotVerifySSLHostname"));
                String wsaHeadersProcessingString = endpoint.getAttribute("WSA_Headers_Processing");
                WSAHeaderProcessing wsaHeadersProcessing = null;
                if (!wsaHeadersProcessingString.isEmpty()) {
                    wsaHeadersProcessing = WSAHeaderProcessing.valueOf(wsaHeadersProcessingString);
                }
                NodeList actions = endpoint.getElementsByTagName("Action");
                loadActions(actionPrefix, url, proxyUrl, idcardMaxAge, doNotVerifySSLHostname, wsaHeadersProcessing, actions);
            } catch (MalformedURLException e) {
                log.error("Error in configuration for endpoint: " + e.getMessage());
                throw new ConfigException();
            }
        }
    }

    private void loadActions(String actionPrefix, URL serviceUrl, URL proxyUrl, int idcardMaxAge, boolean doNotVerifySSLHostname, WSAHeaderProcessing wsaHeadersProcessing, NodeList actions) {
        for (int j = 0; j < actions.getLength(); j++) {
            Route route = new Route();

            Element action = (Element) actions.item(j);
            String actionName = action.getAttribute("Name");
            String soapAction = actionPrefix + actionName;
            route.setSoapAction(soapAction);

            String serviceIdentifier = action.getAttribute("ServiceIdentifier");
            if(!serviceIdentifier.isEmpty()) {
                route.setServiceIdentifier(serviceIdentifier);
            }

            DecouplingInfo info = loadDecouplingInfo(action);
            try {
                DecouplingValidator.validate(info);
                route.setDecouplingInfo(info);
            } catch (DccFault e) {
                log.error("Error in decoupling configuration for action " + actionName + ": " + e.getMessage());
            }

            route.setServiceURL(serviceUrl);
            route.setProxyURL(proxyUrl);
            route.setDoNotVerifySSLHostname(doNotVerifySSLHostname);
            route.setWsaHeaderProcessing(wsaHeadersProcessing);
            route.setIdCardMaxAge(idcardMaxAge);
            routes.add(route);
        }
    }

    private DecouplingInfo loadDecouplingInfo(Element action) {
        DecouplingInfo info = new DecouplingInfo();

        Node timeoutNode = action.getElementsByTagName("Timeout").item(0);
        if (timeoutNode != null) {
            int timeout = Integer.parseInt(timeoutNode.getFirstChild().getNodeValue());
            info.setTimeout(timeout);
        }
        return info;
    }
}