/*
 * SOSI-DCC - Copyright (C) 2011 National Board of e-Health (NSI)
 *
 * All source code and information supplied as part of SOSI-DCC is
 * copyright to National Board of e-Health.
 *
 * The source code has been released under a dual license - meaning you can
 * use either licensed version of the library with your code.
 *
 * It is released under the Common Public License 1.0, a copy of which can
 * be found at the link below.
 * http://www.opensource.org/licenses/cpl1.0.php
 *
 * It is released under the LGPL (GNU Lesser General Public License), either
 * version 2.1 of the License, or (at your option) any later version. A copy
 * of which can be found at the link below.
 * http://www.gnu.org/copyleft/lesser.html
 *
 * $HeadURL: https://svn.nspop.dk/svn/lakeside/sosi-dcc/tags/release-2.4.3/modules/dcc/src/test/java/dk/rsd/dcc/DccConfigTest.java $
 * $Id: DccConfigTest.java 33 2017-01-12 16:52:05Z chg $
 */
package dk.rsd.dcc;

import java.io.StringWriter;
import java.net.URL;
import java.security.KeyStore;

import javax.naming.Context;
import javax.naming.NamingException;

import dk.rsd.dcc.fault.DccFault;
import dk.rsd.dcc.routing.WSAHeaderProcessing;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.apache.log4j.WriterAppender;
import org.h2.jdbcx.JdbcDataSource;
import org.springframework.jdbc.core.JdbcTemplate;

import dk.rsd.dcc.routing.DatabaseTestUtils;
import dk.sosi.seal.model.constants.DGWSConstants;
import junit.framework.TestCase;

import static dk.rsd.dcc.DccConfig.ConfigException;

public class DccConfigTest extends TestCase {

    private StringWriter writer;

    protected void setUp() throws Exception {
        super.setUp();

        // Make sure to reset before any tests are run - safer than teardown as we don't know what happened before this test!!
        writer = new StringWriter();
        BasicConfigurator.resetConfiguration();

        
        PatternLayout patternLayout = new PatternLayout("%-5p %c - %m%n") {

            public boolean ignoresThrowable() {
                return true;
            }

        };
        BasicConfigurator.configure(new WriterAppender(patternLayout, writer));
        Logger.getLogger("org.springframework").setLevel(Level.OFF);
    }

    public void testDccConfig() throws Exception {
        System.setProperty(DccConstants.PROPERTY_DCC_CONFIGFILE, "config/dcc-config-test.xml");
        DccConfig.reload();
        assertLoadingConfiguration();

        String soapAction = "http://www.dkma.dk/medicinecard/xml.schema/2011/01/01#GetMedicineCard";
        assertEquals(DccConfig.getInstance().getServiceURL(soapAction, null), new URL("http://medicinkortet.trifork.com/stub/ws/MedicineCard"));
        assertEquals(DccConfig.getInstance().getProxyURL(soapAction, null), new URL("http://localhost:8080/sosigw/proxy/soap-request"));
        assertEquals(60, DccConfig.getInstance().getIdCardMaxAgeInMinutes(soapAction, null));
        DecouplingInfo info = DccConfig.getInstance().getDefaultDecouplingInfo(soapAction, null);
        assertNotNull(info);
        assertEquals(30000, info.getTimeout());

        soapAction = "http://www.dkma.dk/medicinecard/xml.schema/2008/06/01#GetMedicineCardVersion";
        assertEquals(DccConfig.getInstance().getServiceURL(soapAction, null), new URL("http://medicinkortet.trifork.com/stub/ws/MedicineCard"));
        assertEquals(DccConfig.getInstance().getProxyURL(soapAction, null), new URL("http://localhost:8080/sosigw/proxy/soap-request"));
        info = DccConfig.getInstance().getDefaultDecouplingInfo(soapAction, null);
        assertNotNull(info);
        assertEquals(20000, info.getTimeout());

        soapAction = "http://www.dkma.dk/medicinecard/xml.schema/2008/06/01#UpdateMedicineCard";
        assertEquals(DccConfig.getInstance().getServiceURL(soapAction, null), new URL("http://medicinkortet.trifork.com/stub/ws/MedicineCard"));
        assertEquals(DccConfig.getInstance().getProxyURL(soapAction, null), new URL("http://localhost:8080/sosigw/proxy/soap-request"));
        info = DccConfig.getInstance().getDefaultDecouplingInfo(soapAction, null);
        assertNotNull(info);

        soapAction = "http://www.dkma.dk/medicinecard/xml.schema/2008/06/01#UpdateDrugMedication";
        assertEquals(DccConfig.getInstance().getServiceURL(soapAction, null), new URL("http://medicinkortet.trifork.com/stub/ws/MedicineCard"));
        assertEquals(DccConfig.getInstance().getProxyURL(soapAction, null), new URL("http://localhost:8080/sosigw/proxy/soap-request"));
        info = DccConfig.getInstance().getDefaultDecouplingInfo(soapAction, null);
        assertNotNull(info);
        assertEquals(25000, info.getTimeout());

        soapAction = "http://www.foo.bar#GetMedicineCard";
        assertEquals(DccConfig.getInstance().getServiceURL(soapAction, null), new URL("http://medicinkortet.trifork.com/ws/MedicineCard"));
        assertNull(DccConfig.getInstance().getProxyURL(soapAction, null));
        assertEquals(-1, DccConfig.getInstance().getIdCardMaxAgeInMinutes(soapAction, null));
        info = DccConfig.getInstance().getDefaultDecouplingInfo(soapAction, null);
        assertNotNull(info);
        assertEquals(10000, info.getTimeout());

        soapAction = "http://www.foo.bar#UpdateDrugMedication";
        assertEquals(DccConfig.getInstance().getServiceURL(soapAction, null), new URL("http://medicinkortet.trifork.com/ws/MedicineCard"));
        assertNull(DccConfig.getInstance().getProxyURL(soapAction, null));
        info = DccConfig.getInstance().getDefaultDecouplingInfo(soapAction, null);
        assertNotNull(info);
        assertEquals(5000, info.getTimeout());


        soapAction = "http://www.foo.bar#GetMedicineCardVersion";
        try {
            DccConfig.getInstance().getServiceURL(soapAction, null);
            fail();
        } catch (DccFault fault) {
            //OK
        }

        try {
            DccConfig.getInstance().getDefaultDecouplingInfo(soapAction, null);
            fail();
        } catch (DccFault fault) {
            //OK
        }
        assertNull(DccConfig.getInstance().getProxyURL(soapAction, null));
        assertEquals(-1, DccConfig.getInstance().getIdCardMaxAgeInMinutes(soapAction, null));

        assertEquals(DGWSConstants.VERSION_1_0_1, DccConfig.getInstance().getDGWSVersion());

        assertFalse(DccConfig.getInstance().useIsAlive());
        assertFalse(DccConfig.getInstance().useSLALog());

        KeyStore trustStore = DccConfig.getInstance().getTrustStore();
        assertNotNull(trustStore);
        assertTrue(trustStore.containsAlias("letsencryptroot"));
        assertTrue(trustStore.isCertificateEntry("letsencryptroot"));
    }

    public void testDccConfigClientWhitelisting() throws DccFault {
        System.setProperty(DccConstants.PROPERTY_DCC_CONFIGFILE, "config/dcc-config-whitelist-test.xml");
        DccConfig.reload();
        assertLoadingConfiguration();

        DccConfig.getInstance().checkClientAccessWhitelist("10.0.0.1");
        DccConfig.getInstance().checkClientAccessWhitelist("10.0.0.2");
        try {
            DccConfig.getInstance().checkClientAccessWhitelist("10.0.0.3");
            fail();
        } catch (DccFault fault) {
            //OK
        }
        DccConfig.getInstance().checkClientAccessWhitelist("172.25.18.23");
        DccConfig.getInstance().checkClientAccessWhitelist("172.25.112.34");
        try {
            DccConfig.getInstance().checkClientAccessWhitelist("172.26.18.23");
            fail();
        } catch (DccFault fault) {
            //OK
        }

        assertNull(DccConfig.getInstance().getTrustStore());
    }

    public void testDccConfigWithDB() throws Exception {
        System.setProperty(DccConstants.PROPERTY_DCC_CONFIGFILE, "config/dcc-config-db-test.xml");
        JdbcDataSource ds = new JdbcDataSource();
        Context ctx = DatabaseTestUtils.initAndBindDataSource(ds);

        JdbcTemplate jdbcTemplate = new JdbcTemplate(ds);
        DatabaseTestUtils.populateDatabase(jdbcTemplate);
        DccConfig.reload();

        try {
            DccConfig.getInstance();
        } finally {
            DatabaseTestUtils.resetDatabase(jdbcTemplate);
            DatabaseTestUtils.unBindDataSource(ctx);

            writer.flush();
            assertTrue(writer.toString().contains("Configuration loaded ..."));
        }

        assertTrue(DccConfig.getInstance().useIsAlive());
        assertTrue(DccConfig.getInstance().useSLALog());

        assertEquals(new URL("http://service.url"), DccConfig.getInstance().getServiceURL("http://foo.bar#doIt", null));
        assertEquals(new URL("http://some_proxy.url"), DccConfig.getInstance().getProxyURL("http://foo.bar#doIt", null));
        assertFalse(DccConfig.getInstance().doNotVerifySSLHostname("http://foo.bar#doIt", null));

        assertEquals(new URL("http://another.service.url"), DccConfig.getInstance().getServiceURL("http://foo.bar#doItAgain",null));
        assertNull(DccConfig.getInstance().getProxyURL("http://foo.bar#doItAgain", null));
        assertTrue(DccConfig.getInstance().doNotVerifySSLHostname("http://foo.bar#doItAgain", null));

        try {
            DccConfig.getInstance().getTrustStore();
            fail();
        } catch (DccConfig.ConfigException e) {
            // expected
        }
    }

    public void testDccConfigPreserveHttpHeaders() {
        System.setProperty(DccConstants.PROPERTY_DCC_CONFIGFILE, "config/dcc-config-test.xml");
        DccConfig.reload();
        assertLoadingConfiguration();

        assertTrue(DccConfig.getInstance().preserveHttpHeader("Accept"));
        assertTrue(DccConfig.getInstance().preserveHttpHeader("Accept-Encoding"));
        assertTrue(DccConfig.getInstance().preserveHttpHeader("Accept-encoding"));
        assertTrue(DccConfig.getInstance().preserveHttpHeader("ACCEPT-ENCODING"));
        assertTrue(DccConfig.getInstance().preserveHttpHeader("AccEPT-EncODing"));
        assertFalse(DccConfig.getInstance().preserveHttpHeader("X_NSP_FOO_BAR"));
    }

    public void testDccConfigPreserveHttpHeadersAldreadySetByDCC() {
        System.setProperty(DccConstants.PROPERTY_DCC_CONFIGFILE, "config/dcc-config-http-headers-set-by-dcc.xml");
        DccConfig.reload();
        assertLoadingConfiguration();

        assertTrue(DccConfig.getInstance().preserveHttpHeader("Accept"));
        assertTrue(DccConfig.getInstance().preserveHttpHeader("Accept-Encoding"));
        assertFalse(DccConfig.getInstance().preserveHttpHeader("X_NSP_FOO_BAR"));

        assertFalse(DccConfig.getInstance().preserveHttpHeader("Content-Type"));
        assertFalse(DccConfig.getInstance().preserveHttpHeader("Content-Length"));
        assertFalse(DccConfig.getInstance().preserveHttpHeader("SOAPAction"));
        assertFalse(DccConfig.getInstance().preserveHttpHeader("X-Forwarded-For"));
    }

    public void testDccConfigSSLHostnameVerification() {
        System.setProperty(DccConstants.PROPERTY_DCC_CONFIGFILE, "config/dcc-config-test.xml");
        DccConfig.reload();
        assertLoadingConfiguration();

        assertFalse(DccConfig.getInstance().doNotVerifySSLHostname("http://www.ssl-example-1.org#doIt", null));
        assertTrue(DccConfig.getInstance().doNotVerifySSLHostname("http://www.ssl-example-2.org#doIt", null));
        assertFalse(DccConfig.getInstance().doNotVerifySSLHostname("http://www.ssl-example-3.org#doIt", null));
    }

    public void testDccConfigWSAProcessing() {
        System.setProperty(DccConstants.PROPERTY_DCC_CONFIGFILE, "config/dcc-config-test.xml");
        DccConfig.reload();
        assertLoadingConfiguration();

        assertEquals(WSAHeaderProcessing.WSA_OVERWRITE, DccConfig.getInstance().getWSAHeaderProcessing("http://www.wsa-example-1.org#doIt", null));
        assertEquals(WSAHeaderProcessing.WSA_OVERWRITE, DccConfig.getInstance().getWSAHeaderProcessing("http://www.wsa-example-2.org#doIt", null));
        assertEquals(WSAHeaderProcessing.WSA_OVERWRITE, DccConfig.getInstance().getWSAHeaderProcessing("http://www.wsa-example-3.org#doIt", null));
        assertEquals(WSAHeaderProcessing.WSA_REMOVE, DccConfig.getInstance().getWSAHeaderProcessing("http://www.wsa-example-4.org#doIt", null));
        assertEquals(WSAHeaderProcessing.WSA_OVERWRITE, DccConfig.getInstance().getWSAHeaderProcessing("http://www.wsa-example-5.org#doIt", null));
        assertEquals(WSAHeaderProcessing.WSA_PASSTHROUGH, DccConfig.getInstance().getWSAHeaderProcessing("http://www.wsa-example-6.org#doIt", null));
        assertEquals(WSAHeaderProcessing.WSA_OVERWRITE, DccConfig.getInstance().getWSAHeaderProcessing("http://www.wsa-example-7.org#doIt", null));
        assertEquals(WSAHeaderProcessing.WSA_OVERWRITE, DccConfig.getInstance().getWSAHeaderProcessing("http://www.wsa-example-8.org#doIt", "service-identifier-1"));
    }

    private void assertLoadingConfiguration() {
        try {
            writer.flush();
            String output = writer.toString();
            if (output == null || output.length() == 0) {
            	fail("output from dccconfig is null or empty");
            }
            assertTrue(output.contains("Configuration loaded ..."));
        } catch (ConfigException e) {
        	throw e;
        }
    }

}
