/*
 * SOSI-DCC - Copyright (C) 2011 National Board of e-Health (NSI)
 *
 * All source code and information supplied as part of SOSI-DCC is
 * copyright to National Board of e-Health.
 *
 * The source code has been released under a dual license - meaning you can
 * use either licensed version of the library with your code.
 *
 * It is released under the Common Public License 1.0, a copy of which can
 * be found at the link below.
 * http://www.opensource.org/licenses/cpl1.0.php
 *
 * It is released under the LGPL (GNU Lesser General Public License), either
 * version 2.1 of the License, or (at your option) any later version. A copy
 * of which can be found at the link below.
 * http://www.gnu.org/copyleft/lesser.html
 *
 * $HeadURL: https://svn.nspop.dk/svn/lakeside/sosi-dcc/tags/release-2.4.3/modules/dcc/src/test/java/dk/rsd/dcc/DccHandlerTest.java $
 * $Id: DccHandlerTest.java 33 2017-01-12 16:52:05Z chg $
 */
package dk.rsd.dcc;

import dk.rsd.dcc.routing.WSAHeaderProcessing;
import dk.sosi.seal.model.constants.DGWSConstants;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.varia.NullAppender;
import org.custommonkey.xmlunit.XMLTestCase;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.*;
import java.net.URL;

public class DccHandlerTest extends XMLTestCase {

	private SAXParser parser;
	private String requestStr;
    private String expectedDecoupledStr;

    protected void setUp() throws Exception {
		super.setUp();
		BasicConfigurator.configure(new NullAppender());
		SAXParserFactory factory = SAXParserFactory.newInstance();
		factory.setNamespaceAware(true);
		factory.setFeature("http://xml.org/sax/features/namespaces", true);
		factory.setFeature("http://xml.org/sax/features/namespace-prefixes", true);

		parser = factory.newSAXParser();
		requestStr = readStreamAndRemoveFormatting(this.getClass().getResourceAsStream("/soap/HandlerTestRequest.xml"));
        expectedDecoupledStr = readStreamAndRemoveFormatting(this.getClass().getResourceAsStream("/soap/HandlerTestRequestStripped.xml"));

		System.setProperty(DccConstants.PROPERTY_DCC_CONFIGFILE, "config/dcc-config-test.xml");
	}


	protected void tearDown() throws Exception {
		super.tearDown();
		BasicConfigurator.resetConfiguration();
		parser = null;
		requestStr = null;
	}


	public void testStripRequest() throws Exception {
        DccContext context = initDccContext();
        parseAndAssert(context);
	}

    public void testStripWSA_1_0_Headers() throws Exception {
        DccContext context = initDccContext();
        requestStr = requestStr.replaceAll("<wsa:To>http://foo.dk/bar</wsa:To>","<wsa:To>http://foo.dk/bar</wsa:To><Action xmlns='http://www.w3.org/2005/08/addressing'>foo#bar</Action>");
        parseAndAssert(context);
    }

    public void testWhitespaceHandling() throws Exception {
        DccContext context = initDccContext();

        replaceExpressions("<saml:AttributeValue>2101</saml:AttributeValue>", "<saml:AttributeValue>2101 </saml:AttributeValue>");
        parseAndAssert(context);

        replaceExpressions("<saml:AttributeValue>Hans</saml:AttributeValue>", "<saml:AttributeValue>\nHans\n</saml:AttributeValue>");
        parseAndAssert(context);

        replaceExpressions("<saml:AttributeValue>Dampf</saml:AttributeValue>", "<saml:AttributeValue>\n\tDampf   \n</saml:AttributeValue>");
        parseAndAssert(context);

        replaceExpressions("</saml:Attribute><saml:Attribute", "</saml:Attribute> <saml:Attribute");
        parseAndAssert(context);

        replaceExpressions("</saml:Attribute> <saml:Attribute", "</saml:Attribute>\n \t<saml:Attribute");
        parseAndAssert(context);

        replaceExpressions("><saml:AttributeValue>", "> \t \n <saml:AttributeValue>");
        parseAndAssert(context);

    }

    private void replaceExpressions(String expression, String replacement) {
        requestStr = requestStr.replaceAll(expression, replacement);
        expectedDecoupledStr = expectedDecoupledStr.replaceAll(expression, replacement);
    }

    private DccContext initDccContext() throws Exception {
        DccContext context = new DccContext();
        context.setServiceURL(new URL("http://medicinkortet.trifork.com/stub/ws/MedicineCard"));
        context.setSoapAction("http://www.dkma.dk/medicinecard/xml.schema/2008/06/01#GetMedicineCard");
        context.setWSAHeaderProcessing(WSAHeaderProcessing.WSA_OVERWRITE);
        return context;
    }

    private void parseAndAssert(DccContext context) throws SAXException, IOException {
        parser.parse(new InputSource(new StringReader(requestStr)), new DccHandler(context));
        String xmlString = context.getDecoupledRequestAsString();
        assertXMLEqual(expectedDecoupledStr, xmlString);
        DecouplingInfo info = context.getDecouplingInfo();
        assertEquals(10, info.getTimeout());
        assertEquals("Harmoni/svs", context.getItSystemName());
        assertEquals("AAABFCFUBUPZ7rXq/eRZSVNPU0k=", context.getMessageID());
        assertEquals("1234", context.getFlowID());
        assertEquals(DGWSConstants.VERSION_1_0_1, context.getDgwsVersion());
    }

	public void testIllegalTimeout() throws IOException {
		requestStr = requestStr.replaceAll("<dcc:Timeout>10</dcc:Timeout>", "<dcc:Timeout>10 seconds</dcc:Timeout>");
		try {
			parser.parse(new InputSource(new StringReader(requestStr)), new DccHandler(new DccContext()));
			fail();
		} catch (SAXException e) {
			assertEquals("Unable to parse decoupling timeout as a number: 10 seconds", e.getMessage());
		}
	}

	public void testMissingITSystemName() throws Exception {
		requestStr = requestStr.replaceAll("Name=\"medcom:ITSystemName\"", "Name=\"medcom:SomeOtherProperty\"");
		DccContext context = new DccContext();
		parser.parse(new InputSource(new StringReader(requestStr)), new DccHandler(context));
		assertNull(context.getItSystemName());
	}

    public void testWSAProcessing() throws Exception {
        String requestResource = "/soap/wsa/Request-without-WSA.xml";

        String expectedDecoupledRequestResource = "/soap/wsa/Decoupled-request-with-inserted-WSA.xml";
        parseAndAssertWSAProcessing(requestResource, expectedDecoupledRequestResource, WSAHeaderProcessing.WSA_OVERWRITE);
        expectedDecoupledRequestResource = "/soap/wsa/Decoupled-request-without-WSA.xml";
        parseAndAssertWSAProcessing(requestResource, expectedDecoupledRequestResource, WSAHeaderProcessing.WSA_PASSTHROUGH);
        parseAndAssertWSAProcessing(requestResource, expectedDecoupledRequestResource, WSAHeaderProcessing.WSA_REMOVE);

        requestResource = "/soap/wsa/Request-with-WSA-2004.xml";
        expectedDecoupledRequestResource = "/soap/wsa/Decoupled-request-with-inserted-WSA.xml";
        parseAndAssertWSAProcessing(requestResource, expectedDecoupledRequestResource, WSAHeaderProcessing.WSA_OVERWRITE);
        expectedDecoupledRequestResource = "/soap/wsa/Decoupled-request-without-WSA.xml";
        parseAndAssertWSAProcessing(requestResource, expectedDecoupledRequestResource, WSAHeaderProcessing.WSA_REMOVE);
        expectedDecoupledRequestResource = "/soap/wsa/Decoupled-request-with-retained-WSA-2004.xml";
        parseAndAssertWSAProcessing(requestResource, expectedDecoupledRequestResource, WSAHeaderProcessing.WSA_PASSTHROUGH);

        requestResource = "/soap/wsa/Request-with-WSA-2005.xml";
        expectedDecoupledRequestResource = "/soap/wsa/Decoupled-request-with-inserted-WSA.xml";
        parseAndAssertWSAProcessing(requestResource, expectedDecoupledRequestResource, WSAHeaderProcessing.WSA_OVERWRITE);
        expectedDecoupledRequestResource = "/soap/wsa/Decoupled-request-without-WSA.xml";
        parseAndAssertWSAProcessing(requestResource, expectedDecoupledRequestResource, WSAHeaderProcessing.WSA_REMOVE);
        expectedDecoupledRequestResource = "/soap/wsa/Decoupled-request-with-retained-WSA-2005.xml";
        parseAndAssertWSAProcessing(requestResource, expectedDecoupledRequestResource, WSAHeaderProcessing.WSA_PASSTHROUGH);

        requestResource = "/soap/wsa/Request-with-WSA-2004-and-2005.xml";
        expectedDecoupledRequestResource = "/soap/wsa/Decoupled-request-with-inserted-WSA.xml";
        parseAndAssertWSAProcessing(requestResource, expectedDecoupledRequestResource, WSAHeaderProcessing.WSA_OVERWRITE);
        expectedDecoupledRequestResource = "/soap/wsa/Decoupled-request-without-WSA.xml";
        parseAndAssertWSAProcessing(requestResource, expectedDecoupledRequestResource, WSAHeaderProcessing.WSA_REMOVE);
        expectedDecoupledRequestResource = "/soap/wsa/Decoupled-request-with-retained-WSA-2004-and-2005.xml";
        parseAndAssertWSAProcessing(requestResource, expectedDecoupledRequestResource, WSAHeaderProcessing.WSA_PASSTHROUGH);

    }

    private void parseAndAssertWSAProcessing(String requestResource, String expectedDecoupledRequestResource, WSAHeaderProcessing wsaOverwrite) throws IOException, SAXException {
        DccContext context = new DccContext();
        context.setServiceURL(new URL("http://some.url.com"));
        context.setSoapAction("http://www.dkma.dk/medicinecard/xml.schema/2011/01/01#GetMedicineCard");
        context.setWSAHeaderProcessing(wsaOverwrite);

        String request = readStreamAndRemoveFormatting(this.getClass().getResourceAsStream(requestResource));
        String expectedDecoupledRequest = readStreamAndRemoveFormatting(this.getClass().getResourceAsStream(expectedDecoupledRequestResource));

        parser.parse(new ByteArrayInputStream(request.getBytes("UTF-8")), new DccHandler(context));

        assertXMLEqual(expectedDecoupledRequest, context.getDecoupledRequestAsString());
    }

    private String readStreamAndRemoveFormatting(InputStream is) throws IOException {
    	ByteArrayOutputStream baos = new ByteArrayOutputStream();
    	byte[] buffer = new byte[8192];
		int n;
		while ((n = is.read(buffer, 0, buffer.length)) != -1) {
			baos.write(buffer, 0, n);
		}
		is.close();
		return baos.toString().replaceAll("\\s+", " ").replaceAll(">\\s", ">").replaceAll("\\s<", "<");
    }

}
