/*
 *
 *  * The MIT License
 *  *
 *  * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *  *
 *  * Copyright (C) 2015 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *  *
 *  * Permission is hereby granted, free of charge, to any person obtaining a copy of
 *  * this software and associated documentation files (the "Software"), to deal in
 *  * the Software without restriction, including without limitation the rights to
 *  * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 *  * of the Software, and to permit persons to whom the Software is furnished to do
 *  * so, subject to the following conditions:
 *  *
 *  * The above copyright notice and this permission notice shall be included in all
 *  * copies or substantial portions of the Software.
 *  *
 *  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *  * SOFTWARE.
 *  *
 *  * $HeadURL: https://svn.nspop.dk/svn/lakeside/sosi-dcc/tags/release-2.4.3/modules/dcc/src/test/java/dk/rsd/dcc/DccParserTest.java $
 *  * $Id: DccParserTest.java 33 2017-01-12 16:52:05Z chg $
 *
 */

package dk.rsd.dcc;

import dk.rsd.dcc.fault.SyntaxFault;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.varia.NullAppender;
import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

import java.io.InputStream;

import static org.junit.Assert.assertTrue;

/**
 * Created by chg on 18/08/15.
 */
public class DccParserTest {

    @Rule
    public ExpectedException expectedException = ExpectedException.none();

    @BeforeClass
    public static void loadDccConfig() {
        BasicConfigurator.configure(new NullAppender());
        System.setProperty(DccConstants.PROPERTY_DCC_CONFIGFILE, "config/dcc-config-test.xml");
        DccConfig.getInstance();
    }

    @AfterClass
    public static void reloadDccConfig() {
        DccConfig.reload();
        BasicConfigurator.resetConfiguration();
    }

    @Test
    public void testMultipartContentTypeMissingBoundaryParameter() throws Exception {
        expectedException.expect(SyntaxFault.class);
        expectedException.expectMessage("Missing 'boundary' parameter in Content-Type header");

        DccContext dccContext = new DccContext();
        dccContext.setContentType("multipart/related;start=\"<rootpart*a993c835-67f1-4046-b3e3-f94490086094@example.jaxws.sun.com>\";type=\"application/xop+xml\";start-info=\"text/xml\"");
        InputStream requestAsStream = this.getClass().getResourceAsStream("/soap/mtom/RetrieveDocumentSetRequest-MTOM.txt");

        new DccParser().parse(requestAsStream, dccContext);
    }

    @Test
    public void testMultipartMissingStartingMimeBoundary() throws Exception {
        expectedException.expect(SyntaxFault.class);
        expectedException.expectMessage("Payload in 'multipart/related' request does not begin with expected MIME-boundary: --uuid:a993c835-67f1-4046-b3e3-f94490086094");

        DccContext dccContext = new DccContext();
        dccContext.setContentType("multipart/related;start=\"<rootpart*a993c835-67f1-4046-b3e3-f94490086094@example.jaxws.sun.com>\";type=\"application/xop+xml\";boundary=\"uuid:a993c835-67f1-4046-b3e3-f94490086094\";start-info=\"text/xml\"");
        InputStream requestAsStream = this.getClass().getResourceAsStream("/soap/GetMedicineCardRequest.xml");

        new DccParser().parse(requestAsStream, dccContext);
    }

    @Test
    public void testMultipartMissingSecondMimeBoundary() throws Exception {
        expectedException.expect(SyntaxFault.class);
        expectedException.expectMessage("Payload in 'multipart/related' request is missing MIME-boundary: --MIME_boundary");

        DccContext dccContext = new DccContext();
        dccContext.setContentType("multipart/related;start=\"<rootpart*a993c835-67f1-4046-b3e3-f94490086094@example.jaxws.sun.com>\";type=\"application/xop+xml\";boundary=\"MIME_boundary\";start-info=\"text/xml\"");
        InputStream requestAsStream = this.getClass().getResourceAsStream("/soap/mtom/MultiPartExample-MissingSecondMimeBoundary.txt");

        new DccParser().parse(requestAsStream, dccContext);
    }

    @Test
    public void testMultipartMissingDoubleCRLF() throws Exception {
        expectedException.expect(SyntaxFault.class);
        expectedException.expectMessage("Payload in 'multipart/related' request does not contain expected double CRLF");

        DccContext dccContext = new DccContext();
        dccContext.setContentType("multipart/related;start=\"<rootpart*a993c835-67f1-4046-b3e3-f94490086094@example.jaxws.sun.com>\";type=\"application/xop+xml\";boundary=\"MIME_boundary\";start-info=\"text/xml\"");
        InputStream requestAsStream = this.getClass().getResourceAsStream("/soap/mtom/MultiPartExample-MissingDoubleCRLF.txt");

        new DccParser().parse(requestAsStream, dccContext);
    }

    @Test
    public void testWellformedMultipartRequestWithOnePart() throws Exception {
        DccContext dccContext = new DccContext();
        dccContext.setContentType("multipart/related;start=\"<rootpart*a993c835-67f1-4046-b3e3-f94490086094@example.jaxws.sun.com>\";type=\"application/xop+xml\";boundary=\"uuid:a993c835-67f1-4046-b3e3-f94490086094\";start-info=\"text/xml\"");
        InputStream requestAsStream = this.getClass().getResourceAsStream("/soap/mtom/RetrieveDocumentSetRequest-MTOM.txt");

        new DccParser().parse(requestAsStream, dccContext);
        assertTrue(dccContext.getDecoupledRequestAsString().startsWith("--uuid:a993c835-67f1-4046-b3e3-f94490086094"));
        assertTrue(dccContext.getDecoupledRequestAsString().endsWith("--uuid:a993c835-67f1-4046-b3e3-f94490086094--"));
    }

    @Test
    public void testWellformedMultipartRequestWithMultipleParts() throws Exception {
        DccContext dccContext = new DccContext();
        dccContext.setContentType("multipart/related;start=\"<rootpart*a993c835-67f1-4046-b3e3-f94490086094@example.jaxws.sun.com>\";type=\"application/xop+xml\";boundary=\"MIME_boundary\";start-info=\"text/xml\"");
        InputStream requestAsStream = this.getClass().getResourceAsStream("/soap/mtom/MultiPartExample.txt");

        new DccParser().parse(requestAsStream, dccContext);
        assertTrue(dccContext.getDecoupledRequestAsString().startsWith("--MIME_boundary"));
        assertTrue(dccContext.getDecoupledRequestAsString().endsWith("--MIME_boundary--"));
    }

    @Test
    public void testWellformedMultipartRequestWithBinaryParts() throws Exception {
        DccContext dccContext = new DccContext();
        dccContext.setContentType("multipart/related;start=\"<rootpart*842312eb-e302-47f3-b30e-be765dabdce3@example.jaxws.sun.com>\";type=\"application/xop+xml\";boundary=\"uuid:94c87a46-1358-46a8-9461-8de392820594\";start-info=\"text/xml\"");
        InputStream requestAsStream = this.getClass().getResourceAsStream("/soap/mtom/MultiPartRequest-binary");

        new DccParser().parse(requestAsStream, dccContext);
        assertTrue(dccContext.getDecoupledRequestAsString().startsWith("--uuid:94c87a46-1358-46a8-9461-8de392820594"));
        assertTrue(dccContext.getDecoupledRequestAsString().endsWith("--uuid:94c87a46-1358-46a8-9461-8de392820594--"));
    }
}
