package dk.rsd.dcc;

import com.meterware.httpunit.GetMethodWebRequest;
import com.meterware.httpunit.PostMethodWebRequest;
import com.meterware.httpunit.WebRequest;
import com.meterware.httpunit.WebResponse;
import com.meterware.servletunit.InvocationContext;
import com.meterware.servletunit.ServletRunner;
import com.meterware.servletunit.ServletUnitClient;
import dk.rsd.dcc.dispatch.DispatcherFactory;
import dk.rsd.dcc.routing.WSAHeaderProcessing;
import junit.framework.TestCase;
import org.junit.Ignore;

import javax.servlet.http.HttpServletRequest;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.net.URL;
import java.util.*;

/**
 * @author $LastChangedBy: chg $ $LastChangedDate: 2017-01-12 17:52:05 +0100 (Thu, 12 Jan 2017) $
 * @version $Revision: 33 $
 */
public class DccServletTest extends TestCase {

    private static final String NAME_DOUBLE_HEADER = "TEST-DOUBLE-HEADER";

    private ServletRunner runner;
    private String previousProperty;

    @Override
    protected void setUp() throws Exception {
        super.setUp();
        runner = new ServletRunner();
        runner.registerServlet("decoupling/*", DccServlet.class.getName());
        previousProperty = System.getProperty(DccConstants.PROPERTY_DCC_CONFIGFILE);
        System.setProperty(DccConstants.PROPERTY_DCC_CONFIGFILE, "config/dcc-config-servlet-test.xml");
    }

    @Override
    protected void tearDown() throws Exception {
        super.tearDown();
        runner.shutDown();
        runner = null;
        if (previousProperty != null) System.setProperty(DccConstants.PROPERTY_DCC_CONFIGFILE, previousProperty);
    }


    public void testHttpHeaders() throws Exception {
        PostMethodWebRequest postRequest = new PostMethodWebRequest("http://foo.bar/decoupling");

        postRequest.setHeaderField("SOAPAction", "http://www.example-1.org#doIt");
        DccContext dccContext = invokeInitContext(postRequest);
        assertEquals("http://www.example-1.org#doIt", dccContext.getSoapAction());
        List<Map.Entry<String, String>> preservedHttpHeaders = dccContext.getPreservedHttpHeaders();
        assertEquals(1, preservedHttpHeaders.size()); // HTTP unit sets 'Accept-Encoding' to 'gzip'
        assertEquals("gzip", retrieveHeader(preservedHttpHeaders, "Accept-Encoding").get(0));

        postRequest.setHeaderField("Accept-Encoding", "compress, gzip");
        postRequest.setHeaderField("Accept", "text/xml, multipart/related");
        postRequest.setHeaderField("X-Forwarded-For", "foo");
        dccContext = invokeInitContext(postRequest);
        preservedHttpHeaders = dccContext.getPreservedHttpHeaders();
        assertEquals("http://www.example-1.org#doIt", dccContext.getSoapAction());
        assertEquals(2, preservedHttpHeaders.size());
        assertEquals("compress, gzip", retrieveHeader(preservedHttpHeaders, "Accept-Encoding").get(0));
        assertEquals("text/xml, multipart/related", retrieveHeader(preservedHttpHeaders, "Accept").get(0));
        assertEquals(0, retrieveHeader(preservedHttpHeaders, "X-Forwarded-For").size());

        postRequest.setHeaderField(NAME_DOUBLE_HEADER, "value ignored by unit-test setup");
        dccContext = invokeInitContext(postRequest);
        preservedHttpHeaders = dccContext.getPreservedHttpHeaders();
        assertEquals("http://www.example-1.org#doIt", dccContext.getSoapAction());
        assertEquals(4, preservedHttpHeaders.size());
        assertEquals("compress, gzip", retrieveHeader(preservedHttpHeaders, "Accept-Encoding").get(0));
        assertEquals("text/xml, multipart/related", retrieveHeader(preservedHttpHeaders, "Accept").get(0));
        assertEquals("header1", retrieveHeader(preservedHttpHeaders, NAME_DOUBLE_HEADER).get(0));
        assertEquals("header2", retrieveHeader(preservedHttpHeaders, NAME_DOUBLE_HEADER).get(1));
        assertEquals(0, retrieveHeader(preservedHttpHeaders, "X-Forwarded-For").size());
    }

    public void testSSLHostnameVerification() throws Exception {
        PostMethodWebRequest postRequest = new PostMethodWebRequest("http://foo.bar/decoupling");

        postRequest.setHeaderField("SOAPAction", "http://www.ssl-example-1.org#doIt");
        DccContext dccContext = invokeInitContext(postRequest);
        assertFalse(dccContext.doNotVerifySSLHostname());

        postRequest.setHeaderField("SOAPAction", "http://www.ssl-example-2.org#doIt");
        dccContext = invokeInitContext(postRequest);
        assertTrue(dccContext.doNotVerifySSLHostname());

        postRequest.setHeaderField("SOAPAction", "http://www.ssl-example-3.org#doIt");
        dccContext = invokeInitContext(postRequest);
        assertFalse(dccContext.doNotVerifySSLHostname());
    }

    public void testWSAHeaderProcessing() throws Exception {
        PostMethodWebRequest postRequest = new PostMethodWebRequest("http://foo.bar/decoupling");

        String soapAction = "http://www.wsa-example-1.org#doIt";
        postRequest.setHeaderField("SOAPAction", soapAction);
        DccContext dccContext = invokeInitContext(postRequest);
        assertEquals(WSAHeaderProcessing.WSA_OVERWRITE, dccContext.getWSAHeaderProcessing());

        soapAction = "http://www.wsa-example-4.org#doIt";
        postRequest.setHeaderField("SOAPAction", soapAction);
        dccContext = invokeInitContext(postRequest);
        assertEquals(WSAHeaderProcessing.WSA_REMOVE, dccContext.getWSAHeaderProcessing());

        soapAction = "http://www.wsa-example-5.org#doIt";
        postRequest.setHeaderField("SOAPAction", soapAction);
        dccContext = invokeInitContext(postRequest);
        assertEquals(WSAHeaderProcessing.WSA_OVERWRITE, dccContext.getWSAHeaderProcessing());

    }

    /* -------- Routing tests --------- */

    public void testRoutingNoProxy() throws Exception {
        PostMethodWebRequest postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/124");
        DccContext dccContext;

        postRequest.setHeaderField("SOAPAction", "http://www.example-1.org#doIt");
        dccContext = invokeInitContext(postRequest);
        assertEquals(new URL("http://foobar.com/example-1"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/example-1"), dccContext.getProxyURL());

        postRequest.setHeaderField("SOAPAction", "http://www.example-2.org#doIt");
        dccContext = invokeInitContext(postRequest);
        assertEquals(new URL("http://foobar.com/example-2/"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/example-2/"), dccContext.getProxyURL());
    }

    // Ignore test - SDS-7070
    public void _testRoutingWithServiceIdentifier() throws Exception {
        PostMethodWebRequest postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/service/service-1/124");
        DccContext dccContext;

        postRequest.setHeaderField("SOAPAction", "http://www.example-1.org#doIt");
        dccContext = invokeInitContext(postRequest);

        assertEquals("service-1", dccContext.getServiceIdentifier());
    }

    public void testRoutingWithoutServiceIdentifier_usesSoapAction() throws Exception {
        PostMethodWebRequest postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/service-1/124");
        DccContext dccContext;

        postRequest.setHeaderField("SOAPAction", "http://www.example-1.org#doIt");
        dccContext = invokeInitContext(postRequest);

        assertNull(dccContext.getServiceIdentifier());
    }

    public void testRoutingWithProxy() throws Exception {
        PostMethodWebRequest postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/124");
        DccContext dccContext;

        postRequest.setHeaderField("SOAPAction", "http://www.example-3.org#doIt");
        dccContext = invokeInitContext(postRequest);
        assertEquals(new URL("http://foobar.com/example-3"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/proxy-3"), dccContext.getProxyURL());

        postRequest.setHeaderField("SOAPAction", "http://www.example-4.org#doIt");
        dccContext = invokeInitContext(postRequest);
        assertEquals(new URL("http://foobar.com/example-4/"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/proxy-4/"), dccContext.getProxyURL());
    }

    // Ignore test - SDS-7070
    public void _testRoutingNoProxyWithURLSuffixPreservation() throws Exception {
        DccContext dccContext;
        PostMethodWebRequest postRequest;

        postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/");
        postRequest.setHeaderField("SOAPAction", "http://www.example-5.org#doIt");
        dccContext = invokeInitContext(postRequest);
        assertEquals(new URL("http://foobar.com/example-5/"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/example-5/"), dccContext.getProxyURL());

        postRequest = new PostMethodWebRequest("http://foo.bar/decoupling");
        postRequest.setHeaderField("SOAPAction", "http://www.example-5.org#doIt");
        dccContext = invokeInitContext(postRequest);
        assertEquals(new URL("http://foobar.com/example-5"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/example-5"), dccContext.getProxyURL());

        postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/124");
        postRequest.setHeaderField("SOAPAction", "http://www.example-5.org#doIt");
        dccContext = invokeInitContext(postRequest);
        assertEquals(new URL("http://foobar.com/example-5/124"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/example-5/124"), dccContext.getProxyURL());

        postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/test/foo/bar/345");
        postRequest.setHeaderField("SOAPAction", "http://www.example-5.org#doIt");
        dccContext = invokeInitContext(postRequest);
        assertEquals(new URL("http://foobar.com/example-5/test/foo/bar/345"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/example-5/test/foo/bar/345"), dccContext.getProxyURL());

        postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/service/service-1/foo/bar/345");
        postRequest.setHeaderField("SOAPAction", "http://www.example-5.org#doIt");
        dccContext = invokeInitContext(postRequest);
        assertEquals(new URL("http://foobar.com/example-5/foo/bar/345"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/example-5/foo/bar/345"), dccContext.getProxyURL());
    }

    public void testRoutingWithProxyWithURLSuffixPreservation() throws Exception {
        PostMethodWebRequest postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/124");
        DccContext dccContext;

        postRequest.setHeaderField("SOAPAction", "http://www.example-6.org#doIt");
        dccContext = invokeInitContext(postRequest);
        assertEquals(new URL("http://foobar.com/example-6/124"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/proxy-6/"), dccContext.getProxyURL());

        postRequest.setHeaderField("SOAPAction", "http://www.example-7.org#doIt");
        dccContext = invokeInitContext(postRequest);
        assertEquals(new URL("http://foobar.com/example-7/"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/proxy-7/124"), dccContext.getProxyURL());

        postRequest.setHeaderField("SOAPAction", "http://www.example-8.org#doIt");
        dccContext = invokeInitContext(postRequest);
        assertEquals(new URL("http://foobar.com/example-8/124"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/proxy-8/124"), dccContext.getProxyURL());
    }

    //Should this be supported??
    public void testRoutingGetRequest() throws Exception {
        GetMethodWebRequest getRequest = new GetMethodWebRequest("http://foo.bar/decoupling/124");
        getRequest.setParameter("mode", "fast");
        DccContext dccContext;

        getRequest.setHeaderField("SOAPAction", "http://www.example-1.org#doIt");
        dccContext = invokeInitContext(getRequest);
        assertEquals(new URL("http://foobar.com/example-1"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/example-1"), dccContext.getProxyURL());

        getRequest.setHeaderField("SOAPAction", "http://www.example-8.org#doIt");
        dccContext = invokeInitContext(getRequest);
        assertEquals(new URL("http://foobar.com/example-8/124?mode=fast"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/proxy-8/124?mode=fast"), dccContext.getProxyURL());

        getRequest.setParameter("bar", new String[]{"1", "2"});
        dccContext = invokeInitContext(getRequest);
        assertEquals(new URL("http://foobar.com/example-8/124?bar=1&bar=2&mode=fast"), dccContext.getServiceURL());
        assertEquals(new URL("http://foobar.com/proxy-8/124?bar=1&bar=2&mode=fast"), dccContext.getProxyURL());
    }

    /* -------- Proxy tests --------- */

    public void testRegularRequest() throws Exception {
        InputStream requestAsStream = this.getClass().getResourceAsStream("/soap/GetMedicineCardRequest.xml");
        InputStream responseAsStream = this.getClass().getResourceAsStream("/soap/GetMedicineCardResponse.xml");

        PostMethodWebRequest postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/", requestAsStream, "text/xml; charset=utf-8");
        postRequest.setHeaderField("SOAPAction", "http://www.dkma.dk/medicinecard/xml.schema/2008/06/01#GetMedicineCard");

        MockDispatcherFactory dispatcherFactory = new MockDispatcherFactory(responseAsStream, "text/xml; charset=utf-8");
        WebResponse webResponse = invokePost(postRequest, dispatcherFactory);

        assertEquals(200, webResponse.getResponseCode());
        assertEquals("text/xml", webResponse.getContentType());
        assertEquals("utf-8", webResponse.getCharacterSet());
        assertEquals(DccUtil.consumeStreamToString(this.getClass().getResourceAsStream("/soap/GetMedicineCardResponse.xml")), webResponse.getText());
    }

    public void testBadRequest() throws Exception {
        InputStream requestAsStream = new ByteArrayInputStream("foobar".getBytes());

        PostMethodWebRequest postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/", requestAsStream, "text/xml; charset=utf-8");
        postRequest.setHeaderField("SOAPAction", "http://www.dkma.dk/medicinecard/xml.schema/2008/06/01#GetMedicineCard");

        MockDispatcherFactory dispatcherFactory = new MockDispatcherFactory(null, "text/xml; charset=utf-8");
        WebResponse webResponse = invokePost(postRequest, dispatcherFactory);

        assertEquals(500, webResponse.getResponseCode());
        assertEquals("text/xml", webResponse.getContentType());
        assertEquals("utf-8", webResponse.getCharacterSet());
    }

    public void testMTOMResponse() throws Exception {
        InputStream requestAsStream = this.getClass().getResourceAsStream("/soap/mtom/RetrieveDocumentSetRequest.xml");
        InputStream responseAsStream = this.getClass().getResourceAsStream("/soap/mtom/RetrieveDocumentSetResponse-MTOM.txt");

        PostMethodWebRequest postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/", requestAsStream, "text/xml; charset=utf-8");
        postRequest.setHeaderField("SOAPAction", "urn:ihe:iti:2007:RetrieveDocumentSet");
        postRequest.setHeaderField("Accept", "text/xml, multipart/related, text/html, image/gif, image/jpeg, *; q=.2, */*; q=.2");

        MockDispatcherFactory dispatcherFactory = new MockDispatcherFactory(responseAsStream, "multipart/related; type=\"application/xop+xml\"; boundary=\"uuid:8da04201-4730-4b77-a6d7-adc6fb816c45\"; start=\"<root.message@cxf.apache.org>\"; start-info=\"text/xml\"");
        WebResponse webResponse = invokePost(postRequest, dispatcherFactory);

        assertEquals(200, webResponse.getResponseCode());
        assertEquals("multipart/related", webResponse.getContentType());

        // Set characterSet to UTF-8 to be able to make an equals assertion.
        // A real consumer would handle each part in the MTOM response with its proper encoding (as specified in each part)
        Field characterSetField = WebResponse.class.getDeclaredField("_characterSet");
        characterSetField.setAccessible(true);
        characterSetField.set(webResponse, "utf-8");

        assertEquals(DccUtil.consumeStreamToString(this.getClass().getResourceAsStream("/soap/mtom/RetrieveDocumentSetResponse-MTOM.txt")), webResponse.getText());
    }

    public void testMTOMRequest() throws Exception {
        InputStream requestAsStream = this.getClass().getResourceAsStream("/soap/mtom/RetrieveDocumentSetRequest-MTOM.txt");
        InputStream responseAsStream = this.getClass().getResourceAsStream("/soap/mtom/RetrieveDocumentSetResponse-MTOM.txt");

        PostMethodWebRequest postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/", requestAsStream, "text/xml; charset=utf-8");
        postRequest.setHeaderField("Content-Type", "multipart/related;start=\"<rootpart*a993c835-67f1-4046-b3e3-f94490086094@example.jaxws.sun.com>\";type=\"application/xop+xml\";boundary=\"uuid:a993c835-67f1-4046-b3e3-f94490086094\";start-info=\"text/xml\"");
        postRequest.setHeaderField("SOAPAction", "urn:ihe:iti:2007:RetrieveDocumentSet");
        postRequest.setHeaderField("Accept", "text/xml, multipart/related, text/html, image/gif, image/jpeg, *; q=.2, */*; q=.2");

        MockDispatcherFactory dispatcherFactory = new MockDispatcherFactory(responseAsStream, "multipart/related; type=\"application/xop+xml\"; boundary=\"uuid:8da04201-4730-4b77-a6d7-adc6fb816c45\"; start=\"<root.message@cxf.apache.org>\"; start-info=\"text/xml\"");
        WebResponse webResponse = invokePost(postRequest, dispatcherFactory);

        DccContext dccContext = dispatcherFactory.getLastDccContext();
        assertEquals(dccContext.getContentType(), "multipart/related;start=\"<rootpart*a993c835-67f1-4046-b3e3-f94490086094@example.jaxws.sun.com>\";type=\"application/xop+xml\";boundary=\"uuid:a993c835-67f1-4046-b3e3-f94490086094\";start-info=\"text/xml\"");
        assertTrue(dccContext.getDecoupledRequestAsString().startsWith("--uuid:a993c835-67f1-4046-b3e3-f94490086094"));
        assertTrue(dccContext.getDecoupledRequestAsString().endsWith("--uuid:a993c835-67f1-4046-b3e3-f94490086094--"));
        assertTrue(dccContext.getDecoupledRequestAsString().contains("<S:Header><To xmlns=\"http://schemas.xmlsoap.org/ws/2004/08/addressing\">http://test1.ekstern-test.nspop.dk:8080/ddsrepository</To><Action xmlns=\"http://schemas.xmlsoap.org/ws/2004/08/addressing\">urn:ihe:iti:2007:RetrieveDocumentSet</Action>"));

        assertEquals(200, webResponse.getResponseCode());
        assertEquals("multipart/related", webResponse.getContentType());

        // Set characterSet to UTF-8 to be able to make an equals assertion.
        // A real consumer would handle each part in the MTOM response with its proper encoding (as specified in each part)
        Field characterSetField = WebResponse.class.getDeclaredField("_characterSet");
        characterSetField.setAccessible(true);
        characterSetField.set(webResponse, "utf-8");

        assertEquals(DccUtil.consumeStreamToString(this.getClass().getResourceAsStream("/soap/mtom/RetrieveDocumentSetResponse-MTOM.txt")), webResponse.getText());
    }

    public void testMultiPartExample() throws Exception {
        InputStream requestAsStream = this.getClass().getResourceAsStream("/soap/mtom/MultiPartExample.txt");
        InputStream responseAsStream = this.getClass().getResourceAsStream("/soap/mtom/RetrieveDocumentSetResponse-MTOM.txt");

        PostMethodWebRequest postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/", requestAsStream, "text/xml; charset=utf-8");
        postRequest.setHeaderField("Content-Type", "multipart/related;start=\"<rootpart*a993c835-67f1-4046-b3e3-f94490086094@example.jaxws.sun.com>\";type=\"application/xop+xml\";boundary=\"MIME_boundary\";start-info=\"text/xml\"");
        postRequest.setHeaderField("SOAPAction", "urn:ihe:iti:2007:RetrieveDocumentSet");
        postRequest.setHeaderField("Accept", "text/xml, multipart/related, text/html, image/gif, image/jpeg, *; q=.2, */*; q=.2");

        MockDispatcherFactory dispatcherFactory = new MockDispatcherFactory(responseAsStream, "multipart/related; type=\"application/xop+xml\"; boundary=\"uuid:8da04201-4730-4b77-a6d7-adc6fb816c45\"; start=\"<root.message@cxf.apache.org>\"; start-info=\"text/xml\"");
        WebResponse webResponse = invokePost(postRequest, dispatcherFactory);

        DccContext dccContext = dispatcherFactory.getLastDccContext();
        assertEquals(dccContext.getContentType(), "multipart/related;start=\"<rootpart*a993c835-67f1-4046-b3e3-f94490086094@example.jaxws.sun.com>\";type=\"application/xop+xml\";boundary=\"MIME_boundary\";start-info=\"text/xml\"");
        assertTrue(dccContext.getDecoupledRequestAsString().startsWith("--MIME_boundary"));
        assertTrue(dccContext.getDecoupledRequestAsString().endsWith("--MIME_boundary--"));

        assertEquals(200, webResponse.getResponseCode());
        assertEquals("multipart/related", webResponse.getContentType());
    }

    public void testMultiPartBinaryExample() throws Exception {
        InputStream requestAsStream = this.getClass().getResourceAsStream("/soap/mtom/MultiPartRequest-binary");
        InputStream responseAsStream = this.getClass().getResourceAsStream("/soap/mtom/MultiPartResponse-binary");

        PostMethodWebRequest postRequest = new PostMethodWebRequest("http://foo.bar/decoupling/", requestAsStream, "text/xml; charset=utf-8");
        postRequest.setHeaderField("Content-Type", "multipart/related;start=\"<rootpart*842312eb-e302-47f3-b30e-be765dabdce3@example.jaxws.sun.com>\";type=\"application/xop+xml\";boundary=\"uuid:94c87a46-1358-46a8-9461-8de392820594\";start-info=\"text/xml\"");
        postRequest.setHeaderField("SOAPAction", "http://mtom/MtomServer/downloadImageRequest");
        postRequest.setHeaderField("Accept", "text/xml, multipart/related");

        MockDispatcherFactory dispatcherFactory = new MockDispatcherFactory(responseAsStream, "multipart/related; type=\"application/xop+xml\"; boundary=\"uuid:8da04201-4730-4b77-a6d7-adc6fb816c45\"; start=\"<root.message@cxf.apache.org>\"; start-info=\"text/xml\"");
        WebResponse webResponse = invokePost(postRequest, dispatcherFactory);

        DccContext dccContext = dispatcherFactory.getLastDccContext();
        assertTrue(dccContext.getDecoupledRequestAsString().startsWith("--uuid:94c87a46-1358-46a8-9461-8de392820594"));
        assertTrue(dccContext.getDecoupledRequestAsString().endsWith("--uuid:94c87a46-1358-46a8-9461-8de392820594--"));

        assertEquals(200, webResponse.getResponseCode());
        assertEquals("multipart/related; type=\"application/xop+xml\"; boundary=\"uuid:8da04201-4730-4b77-a6d7-adc6fb816c45\"; start=\"<root.message@cxf.apache.org>\"; start-info=\"text/xml\"", dccContext.getResponseContentType());

        byte[] expectedResponseBytes = DccUtil.consumeStreamToByteArray(this.getClass().getResourceAsStream("/soap/mtom/MultiPartResponse-binary"));
        byte[] actualResponseBytes = DccUtil.consumeStreamToByteArray(webResponse.getInputStream());

        assertTrue(Arrays.equals(expectedResponseBytes, actualResponseBytes));

    }

    /* -------- Helper methods --------- */

    private DccContext invokeInitContext(WebRequest webRequest) throws Exception {
        InvocationContext invocationContext = getNewInvocationContext(webRequest);
        DccServlet dccServlet = (DccServlet) invocationContext.getServlet();
        DccContext dccContext = new DccContext();
        final HttpServletRequest request = invocationContext.getRequest();

        //HttpUnit doesn't support duplicate http headers - use a dynamic proxy to emulate one
        InvocationHandler invocationHandler = new InvocationHandler() {
            public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
                if (method.getName().equals("getHeaders") && NAME_DOUBLE_HEADER.equalsIgnoreCase((String) args[0])) {
                    return new StringTokenizer("header1 header2"); // Quick and dirty 'Enumeration' implementation
                } else {
                    return method.invoke(request, args);
                }
            }
        };

        HttpServletRequest proxyRequest = (HttpServletRequest)
                Proxy.newProxyInstance(this.getClass().getClassLoader(), new Class[]{HttpServletRequest.class}, invocationHandler);

        dccServlet.initContext(proxyRequest, dccContext);
        return dccContext;
    }

    private WebResponse invokePost(WebRequest webRequest, DispatcherFactory dispatcherFactory) throws Exception {
        InvocationContext invocationContext = getNewInvocationContext(webRequest);
        DccServlet dccServlet = (DccServlet) invocationContext.getServlet();
        dccServlet.setDispatcherFactory(dispatcherFactory);
        invocationContext.service();
        return invocationContext.getServletResponse();
    }

    private InvocationContext getNewInvocationContext(WebRequest webRequest) throws Exception {
        ServletUnitClient client = runner.newClient();
        return client.newInvocation(webRequest);
    }

    private List<String> retrieveHeader(Iterable<? extends Map.Entry<String, String>> preservedHttpHeaders, String headerName) {
        List<String> headerValues = new LinkedList<String>();
        for (Map.Entry<String, String> headerEntry : preservedHttpHeaders) {
            if (headerEntry.getKey().equalsIgnoreCase(headerName)) {
                headerValues.add(headerEntry.getValue());
            }
        }
        return headerValues;
    }

}