/*
 * SOSI-DCC - Copyright (C) 2011 National Board of e-Health (NSI)
 *
 * All source code and information supplied as part of SOSI-DCC is
 * copyright to National Board of e-Health.
 *
 * The source code has been released under a dual license - meaning you can
 * use either licensed version of the library with your code.
 *
 * It is released under the Common Public License 1.0, a copy of which can
 * be found at the link below.
 * http://www.opensource.org/licenses/cpl1.0.php
 *
 * It is released under the LGPL (GNU Lesser General Public License), either
 * version 2.1 of the License, or (at your option) any later version. A copy
 * of which can be found at the link below.
 * http://www.gnu.org/copyleft/lesser.html
 *
 * $HeadURL: https://svn.nspop.dk/svn/lakeside/sosi-dcc/tags/release-2.4.3/modules/dcc/src/test/java/dk/rsd/dcc/client/DccTestClient.java $
 * $Id: DccTestClient.java 33 2017-01-12 16:52:05Z chg $
 */
package dk.rsd.dcc.client;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.Reply;
import dk.sosi.seal.vault.EmptyCredentialVault;
import org.custommonkey.xmlunit.XMLTestCase;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URL;

public class DccTestClient extends XMLTestCase {
	private static final String SOAP_ACTION = "http://www.dkma.dk/medicinecard/xml.schema/2008.01.01#GetMedicineCard";
	//private static final String HOST = "localhost:8080";
	private static final String HOST = "192.168.2.2:9080";

	// private static final String HOST = "rsd.lakeside.dk";
	private URL url;
	private String request;

	protected void setUp() throws Exception {
		super.setUp();
		url = new URL("http://" + HOST + "/decoupling/medicinecard/");
		request = readStream(this.getClass().getResourceAsStream("/soap/GetMedicineCardRequest.xml"));
	}

	protected void tearDown() throws Exception {
		super.tearDown();
		url = null;
		request = null;
	}

	public void testClient() throws Exception {
		SoapClient soapClient = new SoapClient();
		String response = readStream(soapClient.callService(url, request, SOAP_ACTION));
		String expectedResponse = readStream(this.getClass().getResourceAsStream("/soap/GetMedicineCardResponse.xml"));
		assertEquals(HttpURLConnection.HTTP_OK, soapClient.getHttpStatus());
		assertXMLEqual(removeFormatting(expectedResponse), removeFormatting(response));
	}

	public void testClientNoDccHeader() throws Exception {
		request = readStream(this.getClass().getResourceAsStream("/soap/GetMedicineCardRequest-NoDccHeader.xml"));

		SoapClient soapClient = new SoapClient();
		String response = readStream(soapClient.callService(url, request, SOAP_ACTION));
		String expectedResponse = readStream(this.getClass().getResourceAsStream("/soap/GetMedicineCardResponse.xml"));
		assertEquals(HttpURLConnection.HTTP_OK, soapClient.getHttpStatus());
		assertXMLEqual(removeFormatting(expectedResponse), removeFormatting(response));
	}

	public void testIllegalTimeout() throws Exception {
		request = request.replaceAll("<dcc:Timeout>5</dcc:Timeout>", "<dcc:Timeout>5 seconds</dcc:Timeout>");

		SoapClient soapClient = new SoapClient();
		String response = readStream(soapClient.callService(url, request, SOAP_ACTION));
		SOSIFactory factory = new SOSIFactory(new EmptyCredentialVault(), System.getProperties());
		Reply reply = factory.deserializeReply(response);
		assertEquals(HttpURLConnection.HTTP_INTERNAL_ERROR, soapClient.getHttpStatus());
		assertTrue(reply.isFault());
		assertEquals("AAABFCFUBUPZ7rXq/eRZSVNPU0k=", reply.getRequestID());
		assertEquals("1234", reply.getFlowID());
		assertEquals("dcc_syntax_error", reply.getFaultCode());
		assertEquals("org.xml.sax.SAXException: Unable to parse decoupling timeout as a number: 5 seconds", reply.getFaultString());
	}

	public void testTimeout() throws Exception {
		request = request.replaceAll("<PersonCivilRegistrationIdentifier>2512484916</PersonCivilRegistrationIdentifier>",
				"<PersonCivilRegistrationIdentifier>sleep 10000</PersonCivilRegistrationIdentifier>");

		SoapClient soapClient = new SoapClient();
		String response = readStream(soapClient.callService(url, request, SOAP_ACTION));
		System.out.println(response);
		SOSIFactory factory = new SOSIFactory(new EmptyCredentialVault(), System.getProperties());
		Reply reply = factory.deserializeReply(response);
		assertEquals(HttpURLConnection.HTTP_INTERNAL_ERROR, soapClient.getHttpStatus());
		assertTrue(reply.isFault());
		assertEquals("dcc_timeout", reply.getFaultCode());
	}

	public void testMixedCaseURL() throws Exception {
		url = new URL("http://" + HOST + "/decoupling/MedicineCard/");

		String response = readStream(new SoapClient().callService(url, request, SOAP_ACTION));
		String expectedResponse = readStream(this.getClass().getResourceAsStream("/soap/GetMedicineCardResponse.xml"));
		assertXMLEqual(removeFormatting(expectedResponse), removeFormatting(response));
	}

	public void testNoIDCardInRequest() throws Exception {
		request = readStream(this.getClass().getResourceAsStream("/soap/GetMedicineCardRequest-NoIDCard.xml"));

		String response = readStream(new SoapClient().callService(url, request, SOAP_ACTION));
		String expectedResponse = readStream(this.getClass().getResourceAsStream("/soap/GetMedicineCardResponse.xml"));
		assertXMLEqual(removeFormatting(expectedResponse), removeFormatting(response));
	}

	public void testSOAPAction() throws Exception {
		request = readStream(this.getClass().getResourceAsStream("/soap/GetMedicineCardRequest-NoDccHeader.xml"));

		String response = readStream(new SoapClient().callService(url, request, "\"" + SOAP_ACTION + "\""));
		String expectedResponse = readStream(this.getClass().getResourceAsStream("/soap/GetMedicineCardResponse.xml"));
		assertXMLEqual(removeFormatting(expectedResponse), removeFormatting(response));

		response = readStream(new SoapClient().callService(url, request, "'" + SOAP_ACTION + "'"));
		expectedResponse = readStream(this.getClass().getResourceAsStream("/soap/GetMedicineCardResponse.xml"));
		assertXMLEqual(removeFormatting(expectedResponse), removeFormatting(response));
	}

	private String readStream(InputStream is) throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		byte[] buffer = new byte[8192];
		int n;

		while ((n = is.read(buffer, 0, buffer.length)) != -1) {
			baos.write(buffer, 0, n);
		}

		return new String(baos.toByteArray(), "UTF-8");
	}

	private String removeFormatting(String xml) {
		return xml.replaceAll("\\s+", " ").replaceAll(">\\s", ">").replaceAll("\\s<", "<");
	}
}
