/*
 *  The MIT License
 *  Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $HeadURL: https://svn.nspop.dk/svn/lakeside/sosi-dcc/tags/release-2.4.3/modules/dcc/src/test/java/dk/rsd/dcc/dispatch/DccServiceCallTest.java $
 * $Id: DccServiceCallTest.java 33 2017-01-12 16:52:05Z chg $
 */

package dk.rsd.dcc.dispatch;

import dk.rsd.dcc.DccConstants;
import dk.rsd.dcc.DccContext;
import dk.rsd.dcc.DccUtil;
import dk.rsd.dcc.dispatch.ssl.DccSSLTestUtils;
import dk.rsd.dcc.dispatch.ssl.SSLDispatcherContext;
import dk.sdsd.dcc.utils.SLALogUtil;
import dk.sdsd.nsp.slalog.api.SLALogger;
import dk.sdsd.nsp.slalog.impl.SLALoggerDummyImpl;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.varia.NullAppender;
import org.eclipse.jetty.http.HttpVersion;
import org.eclipse.jetty.server.*;
import org.eclipse.jetty.server.handler.AbstractHandler;
import org.eclipse.jetty.util.ssl.SslContextFactory;
import org.junit.*;

import javax.net.ssl.SSLHandshakeException;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.security.*;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;

import static junit.framework.TestCase.assertEquals;
import static junit.framework.TestCase.assertTrue;
import static junit.framework.TestCase.fail;

/**
 * Created by chg on 20/04/2016.
 */
public class DccServiceCallTest {

    private static final int HTTP_PORT = 9080;
    private static final int HTTPS_PORT = 9443;

    private static final String KEY_STORE_PASSWORD = "password";
    private static final String RESPONSE_PAYLOAD = "blåbærgrød";

    private static String previousDccConfigFileProperty;
    private static SLALogger previousSLALogger;
    
    private Server server;

    private static class DccJettyHandler extends AbstractHandler {
        @Override
        public void handle(String s, Request request, HttpServletRequest httpServletRequest, HttpServletResponse httpServletResponse) throws IOException, ServletException {
            httpServletResponse.setContentType("text/plain; charset=utf-8");
            httpServletResponse.setStatus(HttpServletResponse.SC_OK);
            httpServletResponse.getWriter().print(RESPONSE_PAYLOAD);
            request.setHandled(true);
        }
    }

    @BeforeClass
    public static void setupConfig() throws GeneralSecurityException {
        BasicConfigurator.configure(new NullAppender());

        previousDccConfigFileProperty = System.getProperty(DccConstants.PROPERTY_DCC_CONFIGFILE);
        System.setProperty(DccConstants.PROPERTY_DCC_CONFIGFILE, "config/dcc-config-test.xml");

        previousSLALogger = SLALogUtil.getSLALogger();
        SLALogUtil.setSLALogger(new SLALoggerDummyImpl());
    }

    @AfterClass
    public static void resetConfig() {
        BasicConfigurator.resetConfiguration();

        if (previousDccConfigFileProperty != null) {
            System.setProperty(DccConstants.PROPERTY_DCC_CONFIGFILE, previousDccConfigFileProperty);
        } else {
            System.getProperties().remove(DccConstants.PROPERTY_DCC_CONFIGFILE);
        }

        SLALogUtil.setSLALogger(previousSLALogger);
    }

    @Before
    public void setUp() throws Exception {
        server = new Server();
        SSLDispatcherContext.init(null);
    }

    @After
    public void tearDown() throws Exception {
        server.stop();
    }

    @Test
    public void testHttpCall() throws Exception {
        DccContext context = getNewInitializedDccContext("http://localhost:" + HTTP_PORT);
        initAndStartJetty(null);
        InputStream responseStream = invokeDccServiceCall(context);
        assertEquals(HttpServletResponse.SC_OK, context.getResponseStatusCode());
        assertEquals(RESPONSE_PAYLOAD, DccUtil.consumeStreamToString(responseStream));
    }

    @Test
    public void testOnlineSSLCallTrustedCert() throws Exception {
        DccContext context = getNewInitializedDccContext("https://www.thawte.com");
        invokeDccServiceCall(context);
        assertEquals(HttpServletResponse.SC_OK, context.getResponseStatusCode());
    }

    @Test
    public void testOnlineSSLCallUntrustedCert() throws Exception {
        DccContext context = getNewInitializedDccContext("https://untrusted-root.badssl.com/");
        try {
            invokeDccServiceCall(context);
            fail("Expected SSLHandshakeException");
        } catch (DccWork.ProxyIOException e) {
            assertTrue(e.getCause() instanceof SSLHandshakeException);
        }
    }

    @Test
    public void testSelfIssuedSSLCert() throws Exception {
        DccContext context = getNewInitializedDccContext("https://localhost:" + HTTPS_PORT);
        runTestForSelfIssuedSSLCert(context, "localhost");
    }

    //@Test
    public void testSelfIssuedSSLWildcardCert() throws Exception {
        DccContext context = getNewInitializedDccContext("https://localhost:" + HTTPS_PORT);
        runTestForSelfIssuedSSLCert(context, "*");
    }

    private void runTestForSelfIssuedSSLCert(DccContext context, String hostname) throws Exception {
        KeyStore keyStore = setupNewKeyStore(hostname);
        initAndStartJetty(keyStore);
        try {
            invokeDccServiceCall(context);
            fail("Expected SSLHandshakeException");
        } catch (DccWork.ProxyIOException e) {
            assertTrue(e.getCause() instanceof SSLHandshakeException);
        }

        SSLDispatcherContext.init(keyStore);

        InputStream responseStream = invokeDccServiceCall(context);
        assertEquals(HttpServletResponse.SC_OK, context.getResponseStatusCode());
        assertEquals(RESPONSE_PAYLOAD, DccUtil.consumeStreamToString(responseStream));

        //Truststore with another certificate
        KeyStore anotherTrustStore = setupNewKeyStore(hostname);
        SSLDispatcherContext.init(anotherTrustStore);

        try {
            invokeDccServiceCall(context);
            fail("Expected SSLHandshakeException");
        } catch (DccWork.ProxyIOException e) {
            assertTrue(e.getCause() instanceof SSLHandshakeException);
        }
    }

    //@Test
    public void testSelfIssuedSSLWrongHostnameCert() throws Exception {
        DccContext context = getNewInitializedDccContext("https://localhost:" + HTTPS_PORT);
        KeyStore keyStore = setupNewKeyStore("foobar.dk");
        initAndStartJetty(keyStore);

        assertNoMatchingHostname(context);

        SSLDispatcherContext.init(keyStore);

        assertNoMatchingHostname(context);

        context.setDoNotVerifySSLHostname(true);

        InputStream responseStream = invokeDccServiceCall(context);
        assertEquals(HttpServletResponse.SC_OK, context.getResponseStatusCode());
        assertEquals(RESPONSE_PAYLOAD, DccUtil.consumeStreamToString(responseStream));

    }

    private void assertNoMatchingHostname(DccContext context) throws Exception {
        try {
            invokeDccServiceCall(context);
            fail("Expected SSLHandshakeException");
        } catch (DccWork.ProxyIOException e) {
            assertTrue(e.getCause() instanceof SSLHandshakeException);
            assertEquals("No name matching localhost found", e.getCause().getCause().getMessage());
        }
    }

    private DccContext getNewInitializedDccContext(String endpoint) throws Exception {
        DccContext context = new DccContext();
        context.setServiceURL(new URL(endpoint));
        context.setDecoupledRequest("<ignored>");
        return context;
    }

    private void initAndStartJetty(KeyStore keyStore) throws Exception {

        HttpConfiguration http_config = new HttpConfiguration();
        http_config.setSecurePort(HTTPS_PORT);

        ServerConnector http = new ServerConnector(server, new HttpConnectionFactory(http_config));
        http.setPort(HTTP_PORT);

        HttpConnectionFactory httpsConnectionFactory = initHttpsConnectionFactory(http_config);
        SslConnectionFactory sslConnectionFactory = initSslConnectionFactory(keyStore, KEY_STORE_PASSWORD);

        ServerConnector https = new ServerConnector(server, sslConnectionFactory, httpsConnectionFactory);
        https.setPort(HTTPS_PORT);

        server.setConnectors(new Connector[]{http, https});
        server.setHandler(new DccJettyHandler());
        server.start();
    }

    private KeyStore setupNewKeyStore(String hostname) throws Exception {
        KeyStore keyStore = KeyStore.getInstance(KeyStore.getDefaultType());
        char[] password = KEY_STORE_PASSWORD.toCharArray();
        keyStore.load(null, password);

        KeyPair keyPair = DccSSLTestUtils.newKeyPair();
        X509Certificate certificate = DccSSLTestUtils.newSSLCertificate(keyPair, hostname);
        keyStore.setKeyEntry("alias", keyPair.getPrivate(), password, new Certificate[]{certificate});
        return keyStore;
    }

    private HttpConnectionFactory initHttpsConnectionFactory(HttpConfiguration http_config) {
        HttpConfiguration https_config = new HttpConfiguration(http_config);
        SecureRequestCustomizer src = new SecureRequestCustomizer();
        src.setStsMaxAge(2000);
        src.setStsIncludeSubDomains(true);
        https_config.addCustomizer(src);
        return new HttpConnectionFactory(https_config);
    }

    private SslConnectionFactory initSslConnectionFactory(KeyStore keyStore, String keyStorePassword) {
        SslContextFactory sslContextFactory = new SslContextFactory();
        sslContextFactory.setKeyStore(keyStore);
        sslContextFactory.setKeyStorePassword(keyStorePassword);
        return new SslConnectionFactory(sslContextFactory, HttpVersion.HTTP_1_1.asString());
    }

    private InputStream invokeDccServiceCall(DccContext context) throws Exception {
        DccServiceCall serviceCall = new DccServiceCall(context);
        serviceCall.run();
        return serviceCall.getResponseStream();
    }

}
