/*
 *
 *  * The MIT License
 *  *
 *  * Original work sponsored and donated by National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *  *
 *  * Copyright (C) 2014 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *  *
 *  * Permission is hereby granted, free of charge, to any person obtaining a copy of
 *  * this software and associated documentation files (the "Software"), to deal in
 *  * the Software without restriction, including without limitation the rights to
 *  * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 *  * of the Software, and to permit persons to whom the Software is furnished to do
 *  * so, subject to the following conditions:
 *  *
 *  * The above copyright notice and this permission notice shall be included in all
 *  * copies or substantial portions of the Software.
 *  *
 *  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *  * SOFTWARE.
 *  *
 *  * $HeadURL: https://svn.nspop.dk/svn/lakeside/sosi-dcc/tags/release-2.4.3/modules/dcc/src/test/java/dk/rsd/dcc/routing/DatabaseBasedRoutingConfiguratorTest.java $
 *  * $Id: DatabaseBasedRoutingConfiguratorTest.java 33 2017-01-12 16:52:05Z chg $
 *
 */

package dk.rsd.dcc.routing;

import dk.rsd.dcc.DecouplingInfo;
import dk.rsd.dcc.fault.DccFault;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.varia.NullAppender;
import org.h2.jdbcx.JdbcDataSource;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;
import org.springframework.jdbc.core.JdbcTemplate;

import javax.naming.Context;
import javax.naming.NamingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Date;

import static dk.rsd.dcc.DccConfig.*;
import static org.junit.Assert.*;

public class DatabaseBasedRoutingConfiguratorTest {

    private static final String ACTIONS_SELECT_STATEMENT = "select SOAP_ACTION, SERVICE_IDENTIFIER, SERVICE_URL, TIMEOUT_MILLIS, IDCARD_MAXAGE_MINS, USE_PROXY, DO_NOT_VERIFY_SSL_HOSTNAME, WSA_HEADERS_PROCESSING from DCC_ACTIONS";

    private static Context ctx;
    private static JdbcTemplate jdbcTemplate;

    @BeforeClass
    public static void setupLoggingAndDatabase() throws NamingException {
        BasicConfigurator.configure(new NullAppender());

        JdbcDataSource ds = new JdbcDataSource();
        ctx = DatabaseTestUtils.initAndBindDataSource(ds);

        jdbcTemplate = new JdbcTemplate(ds);
    }

    @AfterClass
    public static void unbindDatabaseAndLogging() throws NamingException {
        DatabaseTestUtils.unBindDataSource(ctx);

        BasicConfigurator.resetConfiguration();
    }

    @After
    public void resetDatabase() {
        DatabaseTestUtils.resetDatabase(jdbcTemplate);
    }

    @Test(expected = ConfigException.class)
    public void testEmptyDataSourceName() {
        new DatabaseBasedRoutingConfigurator("", null, ACTIONS_SELECT_STATEMENT);
    }

    @Test(expected = ConfigException.class)
    public void testUnboundDataSourceName() {
        new DatabaseBasedRoutingConfigurator("jdbc/foo", null, ACTIONS_SELECT_STATEMENT);
    }

    @Test(expected = ConfigException.class)
    public void testMissingTable() {
        DatabaseBasedRoutingConfigurator configurator = new DatabaseBasedRoutingConfigurator("jdbc/dcc_test", null, ACTIONS_SELECT_STATEMENT);
        configurator.getConfiguration();
    }

    @Test(expected = ConfigException.class)
    public void testMalformedWSAHeaderProcessing() {
        DatabaseTestUtils.populateDatabase(jdbcTemplate);
        DatabaseTestUtils.insertMalformedWSAHeaderProcessing(jdbcTemplate);

        DatabaseBasedRoutingConfigurator configurator = new DatabaseBasedRoutingConfigurator("jdbc/dcc_test", null, ACTIONS_SELECT_STATEMENT);
        configurator.getConfiguration();
    }


    @Test
    public void testValidConfiguration() throws Exception {

        DatabaseTestUtils.populateDatabase(jdbcTemplate);

        DatabaseBasedRoutingConfigurator configurator = new DatabaseBasedRoutingConfigurator("jdbc/dcc_test", new URL("http://proxy.url"), ACTIONS_SELECT_STATEMENT);
        RoutingConfiguration conf = configurator.getConfiguration();
        assertNotNull(conf);

        String soapAction = "http://foo.bar#doIt";
        assertEquals(new URL("http://service.url"), conf.getServiceURL(soapAction, null));
        assertEquals(new URL("http://proxy.url"), conf.getProxyURL(soapAction, null));
        assertEquals(60, conf.getIdCardMaxAgeInMinutes(soapAction, null));
        DecouplingInfo info = conf.getDefaultDecouplingInfo(soapAction, null);
        assertEquals(10000, info.getTimeout());
        assertFalse(conf.doNotVerifySSLHostname(soapAction, null));
        assertEquals(WSAHeaderProcessing.WSA_OVERWRITE, conf.getWSAHeaderProcessing(soapAction, null));

        info = conf.getDefaultDecouplingInfo(soapAction, null);
        assertEquals(10000, info.getTimeout());

        soapAction = "http://foo.bar#doItAgain";
        assertEquals(new URL("http://another.service.url"), conf.getServiceURL(soapAction, null));
        assertNull(conf.getProxyURL(soapAction, null));
        assertEquals(-1, conf.getIdCardMaxAgeInMinutes(soapAction, null));
        info = conf.getDefaultDecouplingInfo(soapAction, null);
        assertEquals(5000, info.getTimeout());
        assertTrue(conf.doNotVerifySSLHostname(soapAction, null));
        assertEquals(WSAHeaderProcessing.WSA_REMOVE, conf.getWSAHeaderProcessing(soapAction, null));

        soapAction = "http://foo.bar#doItOneMoreTime";
        assertEquals(new URL("http://another.service.url"), conf.getServiceURL(soapAction, null));
        assertNull(conf.getProxyURL(soapAction, null));
        assertEquals(-1, conf.getIdCardMaxAgeInMinutes(soapAction, null));
        info = conf.getDefaultDecouplingInfo(soapAction, null);
        assertEquals(5000, info.getTimeout());
        assertFalse(conf.doNotVerifySSLHostname(soapAction, null));
        assertEquals(WSAHeaderProcessing.WSA_PASSTHROUGH, conf.getWSAHeaderProcessing(soapAction, null));

        soapAction = "http://foo.bar#doItYetOneMoreTime";
        assertEquals(WSAHeaderProcessing.WSA_OVERWRITE, conf.getWSAHeaderProcessing(soapAction, null));
    }

    @Test
    public void testIdCardMaxAgeHandling() throws MalformedURLException, DccFault {
        DatabaseTestUtils.populateDatabase(jdbcTemplate);

        DatabaseBasedRoutingConfigurator configurator = new DatabaseBasedRoutingConfigurator("jdbc/dcc_test", new URL("http://proxy.url"), ACTIONS_SELECT_STATEMENT);
        RoutingConfiguration conf = configurator.getConfiguration();

        String soapAction = "http://foo.bar#doIt";
        assertEquals(new URL("http://proxy.url"), conf.getProxyURL(soapAction, null));
        assertEquals(60, conf.getIdCardMaxAgeInMinutes(soapAction, null));

        soapAction = "http://foo.bar#doItAgain";
        assertNull(conf.getProxyURL(soapAction, null));
        assertEquals(-1, conf.getIdCardMaxAgeInMinutes(soapAction, null));

        soapAction = "http://foo.bar#doItOneMoreTime";
        assertNull(conf.getProxyURL(soapAction, null));
        assertEquals(-1, conf.getIdCardMaxAgeInMinutes(soapAction, null));

        configurator = new DatabaseBasedRoutingConfigurator("jdbc/dcc_test", null, ACTIONS_SELECT_STATEMENT);
        conf = configurator.getConfiguration();

        soapAction = "http://foo.bar#doIt";
        assertNull(conf.getProxyURL(soapAction, null));
        assertEquals(-1, conf.getIdCardMaxAgeInMinutes(soapAction, null));

        soapAction = "http://foo.bar#doItAgain";
        assertNull(conf.getProxyURL(soapAction, null));
        assertEquals(-1, conf.getIdCardMaxAgeInMinutes(soapAction, null));

        soapAction = "http://foo.bar#doItOneMoreTime";
        assertNull(conf.getProxyURL(soapAction, null));
        assertEquals(-1, conf.getIdCardMaxAgeInMinutes(soapAction, null));

    }
}